from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, cast

from typing_extensions import Self

from cognite.client.data_classes._base import CogniteObject, CogniteResource, CogniteResourceList
from cognite.client.utils._text import convert_all_keys_recursive

if TYPE_CHECKING:
    from cognite.client import CogniteClient


class TransformationSchemaType(CogniteObject):
    def __init__(self, type: str | None = None) -> None:
        self.type = type


class TransformationSchemaArrayType(TransformationSchemaType):
    def __init__(self, type: str | None = None, element_type: str | None = None, contains_null: bool = False) -> None:
        super().__init__(type=type)
        self.element_type = element_type
        self.contains_null = contains_null


class TransformationSchemaStructType(TransformationSchemaType):
    # TODO: Fields should probably be translated into an object
    def __init__(self, type: str | None = None, fields: list[dict[str, Any]] | None = None) -> None:
        super().__init__(type=type)
        self.fields = fields

    def dump(self, camel_case: bool = True) -> dict:
        dumped = super().dump(camel_case=camel_case)
        return convert_all_keys_recursive(dumped, camel_case=camel_case)  # <-- 'fields' is a nested object


class TransformationSchemaMapType(TransformationSchemaType):
    def __init__(
        self,
        type: str,
        key_type: str | None = None,
        value_type: str | None = None,
        value_contains_null: bool = False,
    ) -> None:
        super().__init__(type=type)
        self.key_type = key_type
        self.value_type = value_type
        self.value_contains_null = value_contains_null

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> TransformationSchemaMapType:
        return cls(
            type=resource["type"],
            key_type=resource.get("keyType"),
            value_type=resource.get("valueType"),
            value_contains_null=resource.get("valueContainsNull"),  # type: ignore[arg-type]
        )


class TransformationSchemaUnknownType(TransformationSchemaType):
    def __init__(self, raw_schema: dict[str, Any]) -> None:
        raw_schema = raw_schema.copy()
        super().__init__(type=raw_schema.pop("type"))  # type is required
        self.__raw_schema = raw_schema

    def dump(self, camel_case: Literal[True] = True) -> dict[str, Any]:  # type: ignore [override]
        return {"type": self.type, **self.__raw_schema}

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(raw_schema=resource)


class TransformationSchemaColumn(CogniteResource):
    """Represents a column of the expected sql structure for a destination type.

    Args:
        name (str | None): Column name
        sql_type (str | None): Type of the column in sql format.
        type (TransformationSchemaType | None): Type of the column in json format.
        nullable (bool): Values for the column can be null or not
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        name: str | None = None,
        sql_type: str | None = None,
        type: TransformationSchemaType | None = None,
        nullable: bool = False,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        self.name = name
        self.sql_type = sql_type
        self.type = type
        self.nullable = nullable
        self._cognite_client = cast("CogniteClient", cognite_client)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if isinstance(self.type, TransformationSchemaType):
            output["type"] = self.type.dump(camel_case=camel_case)
        return output

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> TransformationSchemaColumn:
        instance = super()._load(resource, cognite_client)
        if isinstance(instance.type, dict):
            instance.type = {
                "array": TransformationSchemaArrayType,
                "map": TransformationSchemaMapType,
                "struct": TransformationSchemaStructType,
            }.get(instance.type["type"], TransformationSchemaUnknownType)._load(instance.type)

        elif isinstance(instance.type, str):
            instance.type = TransformationSchemaType(type=instance.type)
        return instance


class TransformationSchemaColumnList(CogniteResourceList[TransformationSchemaColumn]):
    _RESOURCE = TransformationSchemaColumn



from __future__ import annotations

RESERVED_EXTERNAL_IDS = frozenset(
    {
        "Query",
        "Mutation",
        "Subscription",
        "String",
        "Int32",
        "Int64",
        "Int",
        "Float32",
        "Float64",
        "Float",
        "Timestamp",
        "JSONObject",
        "Date",
        "Numeric",
        "Boolean",
        "PageInfo",
        "File",
        "Sequence",
        "TimeSeries",
    }
)
RESERVED_SPACE_IDS = frozenset({"space", "cdf", "dms", "pg3", "shared", "system", "node", "edge"})

RESERVED_PROPERTIES = frozenset(
    {
        "space",
        "externalId",
        "createdTime",
        "lastUpdatedTime",
        "deletedTime",
        "edge_id",
        "node_id",
        "project_id",
        "property_group",
        "seq",
        "tg_table_name",
        "extensions",
    }
)


def validate_data_modeling_identifier(space: str | None, external_id: str | None = None) -> None:
    if space and space in RESERVED_SPACE_IDS:
        raise ValueError(f"The space ID: {space} is reserved. Please use another ID.")
    if external_id and external_id in RESERVED_EXTERNAL_IDS:
        raise ValueError(f"The external ID: {external_id} is reserved. Please use another ID.")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from cognite.client.data_classes._base import (
    CogniteFilter,
    CogniteLabelUpdate,
    CogniteListUpdate,
    CogniteObjectUpdate,
    CognitePrimitiveUpdate,
    CogniteResource,
    CogniteResourceList,
    CogniteUpdate,
    ExternalIDTransformerMixin,
    IdTransformerMixin,
    PropertySpec,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.shared import TimestampRange

if TYPE_CHECKING:
    from cognite.client import CogniteClient


class DataSetCore(WriteableCogniteResource["DataSetWrite"]):
    """Data sets let you document and track data lineage, ensure data integrity, and allow 3rd parties to write their insights securely back to a Cognite Data Fusion (CDF) project.
    This is the read version of the DataSet, which is used when retrieving from CDF.

    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        name (str | None): The name of the data set.
        description (str | None): The description of the data set.
        metadata (dict[str, str] | None): Custom, application-specific metadata. String key -> String value. Limits: Maximum length of key is 128 bytes, value 10240 bytes, up to 256 key-value pairs, of total size at most 10240.
        write_protected (bool | None): To write data to a write-protected data set, you need to be a member of a group that has the "datasets:owner" action for the data set. To learn more about write-protected data sets, follow this [guide](/cdf/data_governance/concepts/datasets/#write-protection)
    """

    def __init__(
        self,
        external_id: str | None = None,
        name: str | None = None,
        description: str | None = None,
        metadata: dict[str, str] | None = None,
        write_protected: bool | None = None,
    ) -> None:
        self.external_id = external_id
        self.name = name
        self.description = description
        self.metadata = metadata
        self.write_protected = write_protected


class DataSet(DataSetCore):
    """Data sets let you document and track data lineage, ensure data integrity, and allow 3rd parties to write their insights securely back to a Cognite Data Fusion (CDF) project.
    This is the read version of the DataSet, which is used when retrieving from CDF.

    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        name (str | None): The name of the data set.
        description (str | None): The description of the data set.
        metadata (dict[str, str] | None): Custom, application-specific metadata. String key -> String value. Limits: Maximum length of key is 128 bytes, value 10240 bytes, up to 256 key-value pairs, of total size at most 10240.
        write_protected (bool | None): To write data to a write-protected data set, you need to be a member of a group that has the "datasets:owner" action for the data set. To learn more about write-protected data sets, follow this [guide](/cdf/data_governance/concepts/datasets/#write-protection)
        id (int | None): A server-generated ID for the object.
        created_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        external_id: str | None = None,
        name: str | None = None,
        description: str | None = None,
        metadata: dict[str, str] | None = None,
        write_protected: bool | None = None,
        id: int | None = None,
        created_time: int | None = None,
        last_updated_time: int | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            name=name,
            description=description,
            metadata=metadata,
            write_protected=write_protected,
        )
        # id/created_time/last_updated_time are required when using the class to read,
        # but don't make sense passing in when creating a new object. So in order to make the typing
        # correct here (i.e. int and not Optional[int]), we force the type to be int rather than
        # Optional[int].
        # TODO: In the next major version we can make these properties required in the constructor
        self.id: int = id  # type: ignore
        self.created_time: int = created_time  # type: ignore
        self.last_updated_time: int = last_updated_time  # type: ignore
        self._cognite_client = cast("CogniteClient", cognite_client)

    def as_write(self) -> DataSetWrite:
        return DataSetWrite(
            external_id=self.external_id,
            name=self.name,
            description=self.description,
            metadata=self.metadata,
            write_protected=self.write_protected,
        )


class DataSetWrite(DataSetCore):
    """Data sets let you document and track data lineage, ensure data integrity, and allow 3rd parties to write their insights securely back to a Cognite Data Fusion (CDF) project.
    This is the read version of the DataSet, which is used when retrieving from CDF.

    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        name (str | None): The name of the data set.
        description (str | None): The description of the data set.
        metadata (dict[str, str] | None): Custom, application-specific metadata. String key -> String value. Limits: Maximum length of key is 128 bytes, value 10240 bytes, up to 256 key-value pairs, of total size at most 10240.
        write_protected (bool | None): To write data to a write-protected data set, you need to be a member of a group that has the "datasets:owner" action for the data set. To learn more about write-protected data sets, follow this [guide](/cdf/data_governance/concepts/datasets/#write-protection)
    """

    def __init__(
        self,
        external_id: str | None = None,
        name: str | None = None,
        description: str | None = None,
        metadata: dict[str, str] | None = None,
        write_protected: bool | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            name=name,
            description=description,
            metadata=metadata,
            write_protected=write_protected,
        )

    def as_write(self) -> DataSetWrite:
        """Returns this DataSetWrite instance."""
        return self


class DataSetFilter(CogniteFilter):
    """Filter on data sets with strict matching.

    Args:
        metadata (dict[str, str] | None): Custom, application specific metadata. String key -> String value. Limits: Maximum length of key is 128 bytes, value 10240 bytes, up to 256 key-value pairs, of total size at most 10240.
        created_time (dict[str, Any] | TimestampRange | None): Range between two timestamps.
        last_updated_time (dict[str, Any] | TimestampRange | None): Range between two timestamps.
        external_id_prefix (str | None): Filter by this (case-sensitive) prefix for the external ID.
        write_protected (bool | None): No description.
    """

    def __init__(
        self,
        metadata: dict[str, str] | None = None,
        created_time: dict[str, Any] | TimestampRange | None = None,
        last_updated_time: dict[str, Any] | TimestampRange | None = None,
        external_id_prefix: str | None = None,
        write_protected: bool | None = None,
    ) -> None:
        self.metadata = metadata
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.external_id_prefix = external_id_prefix
        self.write_protected = write_protected

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        dumped = super().dump(camel_case=camel_case)
        if self.created_time and isinstance(self.created_time, TimestampRange):
            dumped["createdTime"] = self.created_time.dump(camel_case=camel_case)
        if self.last_updated_time and isinstance(self.last_updated_time, TimestampRange):
            dumped["lastUpdatedTime"] = self.last_updated_time.dump(camel_case=camel_case)
        return dumped


class DataSetUpdate(CogniteUpdate):
    """Update applied to single data set

    Args:
        id (int): A server-generated ID for the object.
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
    """

    class _PrimitiveDataSetUpdate(CognitePrimitiveUpdate):
        def set(self, value: Any) -> DataSetUpdate:
            return self._set(value)

    class _ObjectDataSetUpdate(CogniteObjectUpdate):
        def set(self, value: dict) -> DataSetUpdate:
            return self._set(value)

        def add(self, value: dict) -> DataSetUpdate:
            return self._add(value)

        def remove(self, value: list) -> DataSetUpdate:
            return self._remove(value)

    class _ListDataSetUpdate(CogniteListUpdate):
        def set(self, value: list) -> DataSetUpdate:
            return self._set(value)

        def add(self, value: list) -> DataSetUpdate:
            return self._add(value)

        def remove(self, value: list) -> DataSetUpdate:
            return self._remove(value)

    class _LabelDataSetUpdate(CogniteLabelUpdate):
        def add(self, value: list) -> DataSetUpdate:
            return self._add(value)

        def remove(self, value: list) -> DataSetUpdate:
            return self._remove(value)

    @property
    def external_id(self) -> _PrimitiveDataSetUpdate:
        return DataSetUpdate._PrimitiveDataSetUpdate(self, "externalId")

    @property
    def name(self) -> _PrimitiveDataSetUpdate:
        return DataSetUpdate._PrimitiveDataSetUpdate(self, "name")

    @property
    def description(self) -> _PrimitiveDataSetUpdate:
        return DataSetUpdate._PrimitiveDataSetUpdate(self, "description")

    @property
    def metadata(self) -> _ObjectDataSetUpdate:
        return DataSetUpdate._ObjectDataSetUpdate(self, "metadata")

    @property
    def write_protected(self) -> _PrimitiveDataSetUpdate:
        return DataSetUpdate._PrimitiveDataSetUpdate(self, "writeProtected")

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            # External ID is nullable, but is used in the upsert logic and thus cannot be nulled out.
            PropertySpec("external_id", is_nullable=False),
            PropertySpec("name"),
            PropertySpec("description"),
            PropertySpec("metadata", is_container=True),
            PropertySpec("write_protected", is_nullable=False),
        ]


class DataSetWriteList(CogniteResourceList[DataSetWrite], ExternalIDTransformerMixin):
    _RESOURCE = DataSetWrite


class DataSetList(WriteableCogniteResourceList[DataSetWrite, DataSet], IdTransformerMixin):
    _RESOURCE = DataSet

    def as_write(self) -> DataSetWriteList:
        return DataSetWriteList([ds.as_write() for ds in self.data], cognite_client=self._get_cognite_client())



from __future__ import annotations

import re
import warnings
from datetime import timezone
from inspect import signature
from itertools import chain
from numbers import Integral
from typing import TYPE_CHECKING, Any, Literal, Sequence

from cognite.client.exceptions import CogniteImportError
from cognite.client.utils._importing import local_import
from cognite.client.utils._text import to_camel_case
from cognite.client.utils._time import TIME_ATTRIBUTES, ZoneInfo

if TYPE_CHECKING:
    import pandas as pd

    from cognite.client.data_classes import DatapointsArrayList, DatapointsList
    from cognite.client.data_classes._base import T_CogniteResource, T_CogniteResourceList


NULLABLE_INT_COLS = {
    "start_time",
    "end_time",
    "asset_id",
    "parent_id",
    "data_set_id",
    "scheduled_time",
    "schedule_id",
    "session_id",
    "deleted_time",
    "last_success",
    "last_failure",
    "last_seen",
    "last_seen_time",
    "last_updated_time",
}
NULLABLE_INT_COLS |= set(map(to_camel_case, NULLABLE_INT_COLS))


def pandas_major_version() -> int:
    from pandas import __version__

    return int(__version__.split(".")[0])


def convert_tz_for_pandas(tz: str | timezone | ZoneInfo | None) -> str | timezone | None:
    if tz is None or isinstance(tz, (str, timezone)):
        return tz
    if isinstance(tz, ZoneInfo):
        # pandas is not happy about ZoneInfo :shrug:
        if tz.key is not None:
            return tz.key
        raise ValueError("timezone of type ZoneInfo does not have the required 'key' attribute set")
    raise ValueError(f"'timezone' not understood, expected one of: [None, datetime.timezone, ZoneInfo], got {type(tz)}")


def concat_dps_dataframe_list(
    dps_lst: DatapointsList | DatapointsArrayList,
    column_names: Literal["id", "external_id"],
    include_aggregate_name: bool,
    include_granularity_name: bool,
    include_status: bool,
) -> pd.DataFrame:
    pd = local_import("pandas")
    dfs = [
        dps.to_pandas(
            column_names=column_names,
            include_aggregate_name=include_aggregate_name,
            include_granularity_name=include_granularity_name,
            include_status=include_status,
        )
        for dps in dps_lst
    ]
    if not dfs:
        return pd.DataFrame(index=pd.to_datetime([]))
    timezones = set(dps.timezone for dps in dps_lst) - {None}
    # If attempting to join naive & aware, pandas will raise (so we don't need to):
    # TypeError: Cannot join tz-naive with tz-aware DatetimeIndex
    if len(timezones) > 1:
        warnings.warn(
            f"When concatenating datapoints localized to different timezones ({sorted(map(str, timezones))}), the "
            "final dataframe index (timestamps) will be a union of the UTC converted timestamps.",
            UserWarning,
        )
    return concat_dataframes_with_nullable_int_cols(dfs)


def notebook_display_with_fallback(inst: T_CogniteResource | T_CogniteResourceList, **kwargs: Any) -> str:
    if "camel_case" in signature(inst.to_pandas).parameters:
        # Default of False enforced (when accepted by method):
        kwargs["camel_case"] = False
    try:
        return inst.to_pandas(**kwargs)._repr_html_()
    except CogniteImportError:
        warnings.warn(
            "The 'cognite-sdk' depends on 'pandas' for pretty-printing objects like 'Asset' or 'DatapointsList' in "
            "(Jupyter) notebooks and similar environments. Consider installing it! Using fallback method.",
            UserWarning,
        )
        return str(inst)


def convert_nullable_int_cols(df: pd.DataFrame) -> pd.DataFrame:
    to_convert = df.columns.intersection(NULLABLE_INT_COLS)
    df[to_convert] = df[to_convert].astype("Int64")
    return df


def convert_timestamp_columns_to_datetime(df: pd.DataFrame) -> pd.DataFrame:
    to_convert = df.columns.intersection(TIME_ATTRIBUTES)
    df[to_convert] = (1_000_000 * df[to_convert]).astype("datetime64[ns]")
    return df


def concat_dataframes_with_nullable_int_cols(dfs: Sequence[pd.DataFrame]) -> pd.DataFrame:
    pd = local_import("pandas")
    int_cols = [
        i for i, dtype in enumerate(chain.from_iterable(df.dtypes for df in dfs)) if issubclass(dtype.type, Integral)
    ]
    # TODO: Performance optimization possible: The more unique each df.index is to the rest of the dfs, the
    # slower `pd.concat` scales. A manual "union(df.index for df in dfs)" + column insertion is faster for large
    # arrays, but require quite a lot of extra code (e.g. dtypes can be str (object), null.int and float)
    df = pd.concat(dfs, axis="columns", sort=True)  # Sort sorts non-concat axis
    if not int_cols:
        return df

    if pandas_major_version() < 2:
        # As of pandas>=1.5.0, converting float cols (that used to be int) to nullable int using iloc raises FutureWarning,
        # but the suggested code change (to use `frame.isetitem(...)`) results in the wrong dtype (object).
        # See Github Issue: https://github.com/pandas-dev/pandas/issues/49922
        with warnings.catch_warnings():
            warnings.filterwarnings(
                action="ignore",
                message=re.escape(
                    "In a future version, `df.iloc[:, i] = newvals` will attempt to set the values inplace"
                ),
                category=FutureWarning,
            )
            df.iloc[:, int_cols] = df.iloc[:, int_cols].astype("Int64")
    else:
        df.isetitem(int_cols, df.iloc[:, int_cols].astype("Int64"))  # They actually fixed it :D
    return df



from __future__ import annotations

from typing import Any

from requests import Response

from cognite.client._api.annotations import AnnotationsAPI
from cognite.client._api.assets import AssetsAPI
from cognite.client._api.data_modeling import DataModelingAPI
from cognite.client._api.data_sets import DataSetsAPI
from cognite.client._api.diagrams import DiagramsAPI
from cognite.client._api.documents import DocumentsAPI
from cognite.client._api.entity_matching import EntityMatchingAPI
from cognite.client._api.events import EventsAPI
from cognite.client._api.extractionpipelines import ExtractionPipelinesAPI
from cognite.client._api.files import FilesAPI
from cognite.client._api.functions import FunctionsAPI
from cognite.client._api.geospatial import GeospatialAPI
from cognite.client._api.iam import IAMAPI
from cognite.client._api.labels import LabelsAPI
from cognite.client._api.raw import RawAPI
from cognite.client._api.relationships import RelationshipsAPI
from cognite.client._api.sequences import SequencesAPI
from cognite.client._api.templates import TemplatesAPI
from cognite.client._api.three_d import ThreeDAPI
from cognite.client._api.time_series import TimeSeriesAPI
from cognite.client._api.transformations import TransformationsAPI
from cognite.client._api.units import UnitAPI
from cognite.client._api.vision import VisionAPI
from cognite.client._api.workflows import WorkflowAPI
from cognite.client._api_client import APIClient
from cognite.client.config import ClientConfig, global_config
from cognite.client.credentials import CredentialProvider, OAuthClientCredentials, OAuthInteractive
from cognite.client.utils._auxiliary import get_current_sdk_version


class CogniteClient:
    """Main entrypoint into Cognite Python SDK.

    All services are made available through this object. See examples below.

    Args:
        config (ClientConfig | None): The configuration for this client.
    """

    _API_VERSION = "v1"

    def __init__(self, config: ClientConfig | None = None) -> None:
        if (client_config := config or global_config.default_client_config) is None:
            raise ValueError(
                "No ClientConfig has been provided, either pass it directly to CogniteClient "
                "or set global_config.default_client_config."
            )
        else:
            self._config = client_config

        # APIs using base_url / resource path:
        self.assets = AssetsAPI(self._config, self._API_VERSION, self)
        self.events = EventsAPI(self._config, self._API_VERSION, self)
        self.files = FilesAPI(self._config, self._API_VERSION, self)
        self.iam = IAMAPI(self._config, self._API_VERSION, self)
        self.data_sets = DataSetsAPI(self._config, self._API_VERSION, self)
        self.sequences = SequencesAPI(self._config, self._API_VERSION, self)
        self.time_series = TimeSeriesAPI(self._config, self._API_VERSION, self)
        self.geospatial = GeospatialAPI(self._config, self._API_VERSION, self)
        self.raw = RawAPI(self._config, self._API_VERSION, self)
        self.three_d = ThreeDAPI(self._config, self._API_VERSION, self)
        self.labels = LabelsAPI(self._config, self._API_VERSION, self)
        self.relationships = RelationshipsAPI(self._config, self._API_VERSION, self)
        self.entity_matching = EntityMatchingAPI(self._config, self._API_VERSION, self)
        self.templates = TemplatesAPI(self._config, self._API_VERSION, self)
        self.vision = VisionAPI(self._config, self._API_VERSION, self)
        self.extraction_pipelines = ExtractionPipelinesAPI(self._config, self._API_VERSION, self)
        self.transformations = TransformationsAPI(self._config, self._API_VERSION, self)
        self.diagrams = DiagramsAPI(self._config, self._API_VERSION, self)
        self.annotations = AnnotationsAPI(self._config, self._API_VERSION, self)
        self.functions = FunctionsAPI(self._config, self._API_VERSION, self)
        self.data_modeling = DataModelingAPI(self._config, self._API_VERSION, self)
        self.documents = DocumentsAPI(self._config, self._API_VERSION, self)
        self.workflows = WorkflowAPI(self._config, self._API_VERSION, self)
        self.units = UnitAPI(self._config, self._API_VERSION, self)
        # APIs just using base_url:
        self._api_client = APIClient(self._config, api_version=None, cognite_client=self)

    def get(self, url: str, params: dict[str, Any] | None = None, headers: dict[str, Any] | None = None) -> Response:
        """Perform a GET request to an arbitrary path in the API."""
        return self._api_client._get(url, params=params, headers=headers)

    def post(
        self,
        url: str,
        json: dict[str, Any],
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
    ) -> Response:
        """Perform a POST request to an arbitrary path in the API."""
        return self._api_client._post(url, json=json, params=params, headers=headers)

    def put(self, url: str, json: dict[str, Any] | None = None, headers: dict[str, Any] | None = None) -> Response:
        """Perform a PUT request to an arbitrary path in the API."""
        return self._api_client._put(url, json=json, headers=headers)

    def delete(self, url: str, params: dict[str, Any] | None = None, headers: dict[str, Any] | None = None) -> Response:
        """Perform a DELETE request to an arbitrary path in the API."""
        return self._api_client._delete(url, params=params, headers=headers)

    @property
    def version(self) -> str:
        """Returns the current SDK version.

        Returns:
            str: The current SDK version
        """
        return get_current_sdk_version()

    @property
    def config(self) -> ClientConfig:
        """Returns a config object containing the configuration for the current client.

        Returns:
            ClientConfig: The configuration object.
        """
        return self._config

    @classmethod
    def default(
        cls,
        project: str,
        cdf_cluster: str,
        credentials: CredentialProvider,
        client_name: str | None = None,
    ) -> CogniteClient:
        """
        Create a CogniteClient with default configuration.

        The default configuration creates the URLs based on the project and cluster:

        * Base URL: "https://{cdf_cluster}.cognitedata.com/

        Args:
            project (str): The CDF project.
            cdf_cluster (str): The CDF cluster where the CDF project is located.
            credentials (CredentialProvider): Credentials. e.g. Token, ClientCredentials.
            client_name (str | None): A user-defined name for the client. Used to identify the number of unique applications/scripts running on top of CDF. If this is not set, the getpass.getuser() is used instead, meaning the username you are logged in with is used.

        Returns:
            CogniteClient: A CogniteClient instance with default configurations.
        """
        return cls(ClientConfig.default(project, cdf_cluster, credentials, client_name=client_name))

    @classmethod
    def default_oauth_client_credentials(
        cls,
        project: str,
        cdf_cluster: str,
        tenant_id: str,
        client_id: str,
        client_secret: str,
        client_name: str | None = None,
    ) -> CogniteClient:
        """
        Create a CogniteClient with default configuration using a client credentials flow.

        The default configuration creates the URLs based on the project and cluster:

        * Base URL: "https://{cdf_cluster}.cognitedata.com/
        * Token URL: "https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        * Scopes: [f"https://{cdf_cluster}.cognitedata.com/.default"]

        Args:
            project (str): The CDF project.
            cdf_cluster (str): The CDF cluster where the CDF project is located.
            tenant_id (str): The Azure tenant ID.
            client_id (str): The Azure client ID.
            client_secret (str): The Azure client secret.
            client_name (str | None): A user-defined name for the client. Used to identify the number of unique applications/scripts running on top of CDF. If this is not set, the getpass.getuser() is used instead, meaning the username you are logged in with is used.

        Returns:
            CogniteClient: A CogniteClient instance with default configurations.
        """

        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)

        return cls.default(project, cdf_cluster, credentials, client_name)

    @classmethod
    def default_oauth_interactive(
        cls,
        project: str,
        cdf_cluster: str,
        tenant_id: str,
        client_id: str,
        client_name: str | None = None,
    ) -> CogniteClient:
        """
        Create a CogniteClient with default configuration using the interactive flow.

        The default configuration creates the URLs based on the tenant_id and cluster:

        * Base URL: "https://{cdf_cluster}.cognitedata.com/
        * Authority URL: "https://login.microsoftonline.com/{tenant_id}"
        * Scopes: [f"https://{cdf_cluster}.cognitedata.com/.default"]

        Args:
            project (str): The CDF project.
            cdf_cluster (str): The CDF cluster where the CDF project is located.
            tenant_id (str): The Azure tenant ID.
            client_id (str): The Azure client ID.
            client_name (str | None): A user-defined name for the client. Used to identify the number of unique applications/scripts running on top of CDF. If this is not set, the getpass.getuser() is used instead, meaning the username you are logged in with is used.

        Returns:
            CogniteClient: A CogniteClient instance with default configurations.
        """
        credentials = OAuthInteractive.default_for_azure_ad(tenant_id, client_id, cdf_cluster)
        return cls.default(project, cdf_cluster, credentials, client_name)



from __future__ import annotations

from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Literal, MutableSequence, Tuple, Union, overload
from urllib.parse import quote

from typing_extensions import TypeAlias

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.workflows import (
    Workflow,
    WorkflowExecution,
    WorkflowExecutionDetailed,
    WorkflowExecutionList,
    WorkflowIds,
    WorkflowList,
    WorkflowStatus,
    WorkflowTaskExecution,
    WorkflowUpsert,
    WorkflowVersion,
    WorkflowVersionId,
    WorkflowVersionList,
    WorkflowVersionUpsert,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils._identifier import (
    IdentifierSequence,
    WorkflowVersionIdentifierSequence,
)
from cognite.client.utils._session import create_session_and_return_nonce
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import ClientConfig, CogniteClient
    from cognite.client.data_classes import ClientCredentials

WorkflowIdentifier: TypeAlias = Union[WorkflowVersionId, Tuple[str, str], str]
WorkflowVersionIdentifier: TypeAlias = Union[WorkflowVersionId, Tuple[str, str]]


def wrap_workflow_ids(
    workflow_version_ids: WorkflowIdentifier | MutableSequence[WorkflowIdentifier] | None,
) -> list[dict[str, Any]]:
    if workflow_version_ids is None:
        return []
    return WorkflowIds.load(workflow_version_ids).dump(camel_case=True, as_external_id=True)


class WorkflowTaskAPI(APIClient):
    _RESOURCE_PATH = "/workflows/tasks"

    def update(
        self, task_id: str, status: Literal["completed", "failed"], output: dict | None = None
    ) -> WorkflowTaskExecution:
        """`Update status of async task. <https://api-docs.cognite.com/20230101/tag/Tasks/operation/UpdateTaskStatus>`_

        For tasks that has been marked with 'is_async = True', the status must be updated by calling this endpoint with either 'completed' or 'failed'.

        Args:
            task_id (str): The server-generated id of the task.
            status (Literal["completed", "failed"]): The new status of the task. Must be either 'completed' or 'failed'.
            output (dict | None): The output of the task. This will be available for tasks that has specified it as an output with the string "${<taskExternalId>.output}"

        Returns:
            WorkflowTaskExecution: The updated task execution.

        Examples:

            Update task with UUID '000560bc-9080-4286-b242-a27bb4819253' to status 'completed':

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.tasks.update("000560bc-9080-4286-b242-a27bb4819253", "completed")

            Update task with UUID '000560bc-9080-4286-b242-a27bb4819253' to status 'failed' with output '{"a": 1, "b": 2}':

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.tasks.update("000560bc-9080-4286-b242-a27bb4819253", "failed", output={"a": 1, "b": 2})

            Trigger workflow, retrieve detailed task execution and update status of the second task (assumed to be async) to 'completed':

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.trigger("my workflow", "1")
                >>> res = client.workflows.executions.retrieve_detailed(res.id)
                >>> res = client.workflows.tasks.update(res.tasks[1].id, "completed")

        """
        body: dict[str, Any] = {"status": status.upper()}
        if output is not None:
            body["output"] = output
        response = self._post(
            url_path=f"{self._RESOURCE_PATH}/{task_id}/update",
            json=body,
        )
        return WorkflowTaskExecution.load(response.json())


class WorkflowExecutionAPI(APIClient):
    _RESOURCE_PATH = "/workflows/executions"

    def retrieve_detailed(self, id: str) -> WorkflowExecutionDetailed | None:
        """`Retrieve a workflow execution with detailed information. <https://api-docs.cognite.com/20230101/tag/Workflow-executions/operation/ExecutionOfSpecificRunOfWorkflow>`_

        Args:
            id (str): The server-generated id of the workflow execution.

        Returns:
            WorkflowExecutionDetailed | None: The requested workflow execution if it exists, None otherwise.

        Examples:

            Retrieve workflow execution with UUID '000560bc-9080-4286-b242-a27bb4819253':

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.retrieve_detailed("000560bc-9080-4286-b242-a27bb4819253")

            List workflow executions and retrieve detailed information for the first one:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.list()
                >>> res = client.workflows.executions.retrieve_detailed(res[0].id)

        """
        try:
            response = self._get(url_path=f"{self._RESOURCE_PATH}/{id}")
        except CogniteAPIError as e:
            if e.code == 400:
                return None
            raise
        return WorkflowExecutionDetailed._load(response.json())

    def trigger(
        self,
        workflow_external_id: str,
        version: str,
        input: dict | None = None,
        metadata: dict | None = None,
        client_credentials: ClientCredentials | None = None,
    ) -> WorkflowExecution:
        """`Trigger a workflow execution. <https://api-docs.cognite.com/20230101/tag/Workflow-executions/operation/TriggerRunOfSpecificVersionOfWorkflow>`_

        Args:
            workflow_external_id (str): External id of the workflow.
            version (str): Version of the workflow.
            input (dict | None): The input to the workflow execution. This will be available for tasks that have specified it as an input with the string "${workflow.input}" See tip below for more information.
            metadata (dict | None): Application specific metadata. Keys have a maximum length of 32 characters, values a maximum of 255, and there can be a maximum of 10 key-value pairs.
            client_credentials (ClientCredentials | None): Specific credentials that should be used to trigger the workflow execution. When passed will take precedence over the current credentials.

        Tip:
            The workflow input can be available in the workflow tasks. For example, if you have a Task with
            function parameters then you can specify it as follows

                >>> from cognite.client.data_classes import WorkflowTask, FunctionTaskParameters
                >>> task = WorkflowTask(
                ...     external_id="my_workflow-task1",
                ...     parameters=FunctionTaskParameters(
                ...         external_id="cdf_deployed_function:my_function",
                ...         data={"workflow_data": "${workflow.input}",}))

        Returns:
            WorkflowExecution: The created workflow execution.

        Examples:

            Trigger a workflow execution for the workflow "foo", version 1:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.trigger("foo", "1")

            Trigger a workflow execution with input data:

                >>> res = client.workflows.executions.trigger("foo", "1", input={"a": 1, "b": 2})

            Trigger a workflow execution using a specific set of client credentials (i.e. not your current credentials):

                >>> import os
                >>> from cognite.client.data_classes import ClientCredentials
                >>> credentials = ClientCredentials("my-client-id", os.environ["MY_CLIENT_SECRET"])
                >>> res = client.workflows.executions.trigger("foo", "1", client_credentials=credentials)
        """
        nonce = create_session_and_return_nonce(
            self._cognite_client, api_name="Workflow API", client_credentials=client_credentials
        )
        body = {"authentication": {"nonce": nonce}}
        if input is not None:
            body["input"] = input
        if metadata is not None:
            body["metadata"] = metadata

        response = self._post(url_path=f"/workflows/{workflow_external_id}/versions/{version}/run", json=body)
        return WorkflowExecution._load(response.json())

    def list(
        self,
        workflow_version_ids: WorkflowVersionIdentifier | MutableSequence[WorkflowVersionIdentifier] | None = None,
        created_time_start: int | None = None,
        created_time_end: int | None = None,
        statuses: WorkflowStatus | MutableSequence[WorkflowStatus] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
    ) -> WorkflowExecutionList:
        """`List workflow executions in the project. <https://api-docs.cognite.com/20230101/tag/Workflow-executions/operation/ListWorkflowExecutions>`_

        Args:
            workflow_version_ids (WorkflowVersionIdentifier | MutableSequence[WorkflowVersionIdentifier] | None): Workflow version id or list of workflow version ids to filter on.
            created_time_start (int | None): Filter out executions that was created before this time. Time is in milliseconds since epoch.
            created_time_end (int | None): Filter out executions that was created after this time. Time is in milliseconds since epoch.
            statuses (WorkflowStatus | MutableSequence[WorkflowStatus] | None): Workflow status or list of workflow statuses to filter on.
            limit (int): Maximum number of results to return. Defaults to 25. Set to -1, float("inf") or None
                        to return all items.
        Returns:
            WorkflowExecutionList: The requested workflow executions.

        Examples:

            Get all workflow executions for workflows 'my_workflow' version '1':

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.list(("my_workflow", "1"))

            Get all workflow executions for workflows after last 24 hours:

                >>> from cognite.client import CogniteClient
                >>> from datetime import datetime, timedelta
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.list(created_time_start=int((datetime.now() - timedelta(days=1)).timestamp() * 1000))

        """
        filter_: dict[str, Any] = {}
        if workflow_version_ids is not None:
            filter_["workflowFilters"] = WorkflowIds.load(workflow_version_ids).dump(
                camel_case=True, as_external_id=True
            )
        if created_time_start is not None:
            filter_["createdTimeStart"] = created_time_start
        if created_time_end is not None:
            filter_["createdTimeEnd"] = created_time_end
        if statuses is not None:
            if isinstance(statuses, MutableSequence):
                filter_["status"] = [status.upper() for status in statuses]
            else:  # Assume it is a stringy type
                filter_["status"] = [statuses.upper()]

        return self._list(
            method="POST",
            resource_cls=WorkflowExecution,
            list_cls=WorkflowExecutionList,
            filter=filter_,
            limit=limit,
        )

    def cancel(self, id: str, reason: str | None) -> WorkflowExecution:
        """`cancel a workflow execution. <https://api-docs.cognite.com/20230101/tag/Workflow-executions/operation/WorkflowExecutionCancellation>`_

        Args:
            id (str): The server-generated id of the workflow execution.
            reason (str | None): The reason for the cancellation, this will be put within the execution's `reasonForIncompletion` field. It is defaulted to 'cancelled' if not provided.

        Note:
            Cancelling a workflow will immediately cancel the `in_progress` tasks, but not their spawned work in
            other services (like transformations and functions).

        Returns:
            WorkflowExecution: The canceled workflow execution.

        Examples:

            Trigger a workflow execution for the workflow "foo", version 1 and cancel it:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.trigger("foo", "1")
                >>> client.workflows.executions.cancel(id="foo", reason="test cancelation")
        """
        response = self._post(
            url_path=f"{self._RESOURCE_PATH}/{id}/cancel",
            json={
                "reason": reason,
            }
            if reason
            else {},
        )

        return WorkflowExecution._load(response.json())

    def retry(self, id: str, client_credentials: ClientCredentials | None = None) -> WorkflowExecution:
        """`Retry a workflow execution. <https://api-docs.cognite.com/20230101/tag/Workflow-executions/operation/WorkflowExecutionRetryn>`_

        Args:
            id (str): The server-generated id of the workflow execution.
            client_credentials (ClientCredentials | None): Specific credentials that should be used to retry the workflow execution. When passed will take precedence over the current credentials.

        Returns:
            WorkflowExecution: The retried workflow execution.

        Examples:
            Retry a workflow execution that has been cancelled or failed:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.executions.trigger("foo", "1")
                >>> client.workflows.executions.cancel(id=res.id, reason="test cancellation")
                >>> client.workflows.executions.retry(res.id)
        """
        nonce = create_session_and_return_nonce(
            self._cognite_client, api_name="Workflow API", client_credentials=client_credentials
        )
        response = self._post(
            url_path=f"{self._RESOURCE_PATH}/{id}/retry",
            json={"authentication": {"nonce": nonce}},
        )
        return WorkflowExecution._load(response.json())


class WorkflowVersionAPI(APIClient):
    _RESOURCE_PATH = "/workflows/versions"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self._DELETE_LIMIT = 100

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        workflow_version_ids: WorkflowIdentifier | MutableSequence[WorkflowIdentifier] | None = None,
        limit: int | None = None,
    ) -> Iterator[WorkflowVersion]: ...

    @overload
    def __call__(
        self,
        chunk_size: int,
        workflow_version_ids: WorkflowIdentifier | MutableSequence[WorkflowIdentifier] | None = None,
        limit: int | None = None,
    ) -> Iterator[WorkflowVersionList]: ...

    def __call__(
        self,
        chunk_size: int | None = None,
        workflow_version_ids: WorkflowIdentifier | MutableSequence[WorkflowIdentifier] | None = None,
        limit: int | None = None,
    ) -> Iterator[WorkflowVersion] | Iterator[WorkflowVersionList]:
        """Iterate over workflow versions

        Args:
            chunk_size (int | None): The number of workflow versions to return in each chunk. Defaults to yielding one workflow version at a time.
            workflow_version_ids (WorkflowIdentifier | MutableSequence[WorkflowIdentifier] | None): Workflow version id or list of workflow version ids to filter on.
            limit (int | None): Maximum number of workflow versions to return. Defaults to returning all.

        Returns:
            Iterator[WorkflowVersion] | Iterator[WorkflowVersionList]: Yields WorkflowVersion one by one if chunk_size is None, otherwise yields WorkflowVersionList objects.
        """
        return self._list_generator(
            method="GET",
            resource_cls=WorkflowVersion,
            list_cls=WorkflowVersionList,
            filter={"workflowFilters": wrap_workflow_ids(workflow_version_ids)},
            limit=limit,
            chunk_size=chunk_size,
        )

    def __iter__(self) -> Iterator[WorkflowVersion]:
        """Iterate all over workflow versions"""
        return self()

    def upsert(self, version: WorkflowVersionUpsert, mode: Literal["replace"] = "replace") -> WorkflowVersion:
        """`Create a workflow version. <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/CreateOrUpdateWorkflowVersion>`_

        Note this is an upsert endpoint, so if a workflow with the same version external id already exists, it will be updated.

        Furthermore, if the workflow does not exist, it will be created.

        Args:
            version (WorkflowVersionUpsert): The workflow version to create or update.
            mode (Literal['replace']): This is not an option for the API, but is included here to document that the upserts are always done in replace mode.

        Returns:
            WorkflowVersion: The created workflow version.

        Examples:

            Create workflow version with one Function task:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import WorkflowVersionUpsert, WorkflowDefinitionUpsert, WorkflowTask, FunctionTaskParameters
                >>> client = CogniteClient()
                >>> new_version = WorkflowVersionUpsert(
                ...    workflow_external_id="my_workflow",
                ...    version="1",
                ...    workflow_definition=WorkflowDefinitionUpsert(
                ...        tasks=[
                ...            WorkflowTask(
                ...                external_id="my_workflow-task1",
                ...                parameters=FunctionTaskParameters(
                ...                    external_id="cdf_deployed_function:my_function",
                ...                    data={"a": 1, "b": 2},
                ...                ),
                ...            )
                ...        ],
                ...        description="This workflow has one step",
                ...    ),
                ... )
                >>> res = client.workflows.versions.upsert(new_version)
        """
        if mode != "replace":
            raise ValueError("Only replace mode is supported for upserting workflow versions.")

        response = self._post(
            url_path=self._RESOURCE_PATH,
            json={"items": [version.dump(camel_case=True)]},
        )

        return WorkflowVersion._load(response.json()["items"][0])

    def delete(
        self,
        workflow_version_id: WorkflowVersionIdentifier | MutableSequence[WorkflowVersionIdentifier],
        ignore_unknown_ids: bool = False,
    ) -> None:
        """`Delete a workflow version(s). <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/DeleteSpecificVersionsOfWorkflow>`_

        Args:
            workflow_version_id (WorkflowVersionIdentifier | MutableSequence[WorkflowVersionIdentifier]): Workflow version id or list of workflow version ids to delete.
            ignore_unknown_ids (bool): Ignore external ids that are not found rather than throw an exception.

        Examples:

            Delete workflow version "1" of workflow "my workflow" specified by using a tuple:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> client.workflows.versions.delete(("my workflow", "1"))

            Delete workflow version "1" of workflow "my workflow" and workflow version "2" of workflow "my workflow 2" using the WorkflowVersionId class:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import WorkflowVersionId
                >>> client = CogniteClient()
                >>> client.workflows.versions.delete([WorkflowVersionId("my workflow", "1"), WorkflowVersionId("my workflow 2", "2")])

        """
        identifiers = WorkflowIds.load(workflow_version_id).dump(camel_case=True)
        self._delete_multiple(
            identifiers=WorkflowVersionIdentifierSequence.load(identifiers),
            params={"ignoreUnknownIds": ignore_unknown_ids},
            wrap_ids=True,
        )

    def retrieve(self, workflow_external_id: str, version: str) -> WorkflowVersion | None:
        """`Retrieve a workflow version. <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/GetSpecificVersion>`_

        Args:
            workflow_external_id (str): External id of the workflow.
            version (str): Version of the workflow.

        Returns:
            WorkflowVersion | None: The requested workflow version if it exists, None otherwise.

        Examples:

            Retrieve workflow version 1 of workflow my workflow:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.versions.retrieve("my workflow", "1")
        """
        try:
            response = self._get(
                url_path=f"/workflows/{quote(workflow_external_id, '')}/versions/{quote(version, '')}",
            )
        except CogniteAPIError as e:
            if e.code == 404:
                return None
            raise e

        return WorkflowVersion._load(response.json())

    def list(
        self,
        workflow_version_ids: WorkflowIdentifier | MutableSequence[WorkflowIdentifier] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
    ) -> WorkflowVersionList:
        """`List workflow versions in the project <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/ListWorkflowVersions>`_

        Args:
            workflow_version_ids (WorkflowIdentifier | MutableSequence[WorkflowIdentifier] | None): Workflow version id or list of workflow version ids to filter on.
            limit (int): Maximum number of results to return. Defaults to 25. Set to -1, float("inf") or None

        Returns:
            WorkflowVersionList: The requested workflow versions.

        Examples:

            Get all workflow version for workflows 'my_workflow' and 'my_workflow_2':

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.versions.list(["my_workflow", "my_workflow_2"])

            Get all workflow versions for workflows 'my_workflow' and 'my_workflow_2' using the WorkflowVersionId class:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import WorkflowVersionId
                >>> client = CogniteClient()
                >>> res = client.workflows.versions.list([WorkflowVersionId("my_workflow"), WorkflowVersionId("my_workflow_2")])

            Get all workflow versions for workflows 'my_workflow' version '1' and 'my_workflow_2' version '2' using tuples:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.versions.list([("my_workflow", "1"), ("my_workflow_2", "2")])

        """
        return self._list(
            method="POST",
            resource_cls=WorkflowVersion,
            list_cls=WorkflowVersionList,
            filter={"workflowFilters": wrap_workflow_ids(workflow_version_ids)},
            limit=limit,
        )


class WorkflowAPI(APIClient):
    _RESOURCE_PATH = "/workflows"

    def __init__(
        self,
        config: ClientConfig,
        api_version: str | None,
        cognite_client: CogniteClient,
    ) -> None:
        super().__init__(config, api_version, cognite_client)
        self.versions = WorkflowVersionAPI(config, api_version, cognite_client)
        self.executions = WorkflowExecutionAPI(config, api_version, cognite_client)
        self.tasks = WorkflowTaskAPI(config, api_version, cognite_client)
        self._DELETE_LIMIT = 100

    @overload
    def __call__(self, chunk_size: None = None, limit: None = None) -> Iterator[Workflow]: ...

    @overload
    def __call__(self, chunk_size: int, limit: None) -> Iterator[Workflow]: ...

    def __call__(
        self, chunk_size: int | None = None, limit: int | None = None
    ) -> Iterator[Workflow] | Iterator[WorkflowList]:
        """Iterate over workflows

        Args:
            chunk_size (int | None): The number of workflows to return in each chunk. Defaults to yielding one workflow at a time.
            limit (int | None): Maximum number of workflows to return. Defaults to returning all items.

        Returns:
            Iterator[Workflow] | Iterator[WorkflowList]: Yields Workflow one by one if chunk_size is None, otherwise yields WorkflowList objects.

        """
        return self._list_generator(
            method="GET", resource_cls=Workflow, list_cls=WorkflowList, limit=limit, chunk_size=chunk_size
        )

    def __iter__(self) -> Iterator[Workflow]:
        """Iterate all over workflows"""
        return self()

    def upsert(self, workflow: WorkflowUpsert, mode: Literal["replace"] = "replace") -> Workflow:
        """`Create a workflow. <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/CreateOrUpdateWorkflow>`_

        Note this is an upsert endpoint, so if a workflow with the same external id already exists, it will be updated.

        Args:
            workflow (WorkflowUpsert): The workflow to create or update.
            mode (Literal['replace']): This is not an option for the API, but is included here to document that the upserts are always done in replace mode.

        Returns:
            Workflow: The created workflow.

        Examples:

            Create workflow my workflow:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import WorkflowUpsert
                >>> client = CogniteClient()
                >>> res = client.workflows.upsert(WorkflowUpsert(external_id="my workflow", description="my workflow description"))
        """
        if mode != "replace":
            raise ValueError("Only replace mode is supported for upserting workflows.")

        response = self._post(
            url_path=self._RESOURCE_PATH,
            json={"items": [workflow.dump(camel_case=True)]},
        )
        return Workflow._load(response.json()["items"][0])

    def retrieve(self, external_id: str) -> Workflow | None:
        """`Retrieve a workflow. <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/CreateOrUpdateWorkflow>`_

        Args:
            external_id (str): Identifier for a Workflow. Must be unique for the project.

        Returns:
            Workflow | None: The requested workflow if it exists, None otherwise.

        Examples:

            Retrieve workflow my workflow:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.retrieve("my workflow")
        """
        try:
            response = self._get(url_path=f"{self._RESOURCE_PATH}/{quote(external_id, '')}")
        except CogniteAPIError as e:
            if e.code == 404:
                return None
            raise e
        return Workflow._load(response.json())

    def delete(self, external_id: str | SequenceNotStr[str], ignore_unknown_ids: bool = False) -> None:
        """`Delete one or more workflows with versions. <https://pr-2282.specs.preview.cogniteapp.com/20230101.json.html#tag/Workflows/operation/DeleteWorkflows>`_

        Args:
            external_id (str | SequenceNotStr[str]): External id or list of external ids to delete.
            ignore_unknown_ids (bool): Ignore external ids that are not found rather than throw an exception.

        Examples:

            Delete workflow my workflow:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> client.workflows.delete("my workflow")
        """
        self._delete_multiple(
            identifiers=IdentifierSequence.load(external_ids=external_id),
            params={"ignoreUnknownIds": ignore_unknown_ids},
            wrap_ids=True,
        )

    def list(self, limit: int | None = DEFAULT_LIMIT_READ) -> WorkflowList:
        """`List all workflows in the project. <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/FetchAllWorkflows>`_

        Args:
            limit (int | None): Maximum number of results to return. Defaults to 25. Set to -1, float("inf") or None
        Returns:
            WorkflowList: All workflows in the CDF project.

        Examples:

            List all workflows:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.list()
        """
        return self._list(
            method="GET",
            resource_cls=Workflow,
            list_cls=WorkflowList,
            limit=limit,
        )



from __future__ import annotations

import inspect
from abc import ABC
from collections.abc import Iterable
from datetime import date, datetime
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Self

from cognite.client.data_classes.data_modeling.data_types import (
    DirectRelationReference,
)
from cognite.client.data_classes.data_modeling.ids import ContainerId, EdgeId, NodeId, ViewId
from cognite.client.data_classes.data_modeling.instances import (
    Edge,
    EdgeApply,
    Node,
    NodeApply,
    _serialize_property_value,
)
from cognite.client.utils._text import to_camel_case

if TYPE_CHECKING:
    from cognite.client import CogniteClient


class PropertyOptions:
    """This is a descriptor class for instance properties in a typed class.

    It is used when you have a property that has a different name in the Data Model
    compared to the name in the Python class.

    Args:
        identifier (str | None): The name of the property in the Data Model. Defaults to the name of the property in the Python class.
    """

    def __init__(self, identifier: str | None = None) -> None:
        self.name = identifier

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = self.name or name
        if self.name in _RESERVED_PROPERTY_NAMES:
            self.name = f"__{self.name}"

    def __get__(self, instance: Any, owner: type) -> Any:
        try:
            return instance.__dict__[self.name]
        except KeyError:
            raise AttributeError(f"'{owner.__name__}' object has no attribute '{self.name}'")

    def __set__(self, instance: Any, value: Any) -> None:
        try:
            instance.__dict__[self.name] = value
        except KeyError:
            raise AttributeError(f"'{instance.__class__.__name__}' object has no attribute '{self.name}'")

    def __delete__(self, instance: Any) -> None:
        try:
            del instance.__dict__[self.name]
        except KeyError:
            raise AttributeError(f"'{instance.__class__.__name__}' object has no attribute '{self.name}'")


class TypedNodeApply(NodeApply, ABC):
    _instance_properties: frozenset[str] = frozenset(
        {"space", "external_id", "existing_version", "type", "instance_type", "sources"}
    )

    @classmethod
    def get_source(cls) -> ContainerId | ViewId:
        raise NotImplementedError

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if "sources" not in output:
            output["sources"] = []
        output["sources"].append(
            {
                "source": self.get_source().dump(camel_case),
                "properties": _dump_properties(self, camel_case, self._instance_properties),
            }
        )
        return output

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        sources = resource.pop("sources", [])
        properties = sources[0].get("properties", {}) if sources else {}
        return cast(Self, _load_instance(cls, resource, properties, cls._instance_properties))


class TypedEdgeApply(EdgeApply, ABC):
    _instance_properties = frozenset(
        {"space", "external_id", "existing_version", "type", "start_node", "end_node", "instance_type", "sources"}
    )

    @classmethod
    def get_source(cls) -> ContainerId | ViewId:
        raise NotImplementedError

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if "sources" not in output:
            output["sources"] = []
        output["sources"].append(
            {
                "source": self.get_source().dump(camel_case),
                "properties": _dump_properties(self, camel_case, self._instance_properties),
            }
        )
        return output

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        sources = resource.pop("sources", [])
        properties = sources[0].get("properties", {}) if sources else {}
        return cast(Self, _load_instance(cls, resource, properties, cls._instance_properties))


class TypedNode(Node, ABC):
    _instance_properties = frozenset(
        {
            "space",
            "external_id",
            "version",
            "last_updated_time",
            "created_time",
            "deleted_time",
            "type",
            "instance_type",
            "properties",
        }
    )

    @classmethod
    def get_source(cls) -> ViewId:
        raise NotImplementedError

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if "properties" not in output:
            output["properties"] = {}
        source = self.get_source()
        output["properties"] = {
            source.space: {
                source.as_source_identifier(): _dump_properties(self, camel_case, self._instance_properties),
            }
        }
        return output

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        all_properties = resource.pop("properties", {})
        source = cls.get_source()
        properties = all_properties.get(source.space, {}).get(source.as_source_identifier(), {})
        return cast(Self, _load_instance(cls, resource, properties, cls._instance_properties))


class TypedEdge(Edge, ABC):
    _instance_properties = frozenset(
        {
            "space",
            "external_id",
            "version",
            "last_updated_time",
            "created_time",
            "deleted_time",
            "type",
            "start_node",
            "end_node",
            "instance_type",
            "properties",
        }
    )

    @classmethod
    def get_source(cls) -> ViewId:
        raise NotImplementedError

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if "properties" not in output:
            output["properties"] = {}
        source = self.get_source()
        output["properties"] = {
            source.space: {
                source.as_source_identifier(): _dump_properties(self, camel_case, self._instance_properties),
            }
        }
        return output

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        all_properties = resource.pop("properties", {})
        source = cls.get_source()
        properties = all_properties.get(source.space, {}).get(source.as_source_identifier(), {})
        return cast(Self, _load_instance(cls, resource, properties, cls._instance_properties))


def _load_instance(
    cls: type, resource: dict[str, Any], properties: dict[str, Any], instance_properties: frozenset[str]
) -> dict[str, Any]:
    args: dict[str, Any] = {}
    resource.pop("instanceType", None)
    signature = inspect.signature(cls.__init__)  # type: ignore[misc]
    args.update(_load_properties(cls, properties, instance_properties, signature))
    args.update(_load_fixed_attributes(resource, instance_properties, signature))
    return cls(**args)


def _load_fixed_attributes(
    resource: dict[str, Any], instance_properties: frozenset[str], signature: inspect.Signature
) -> dict[str, Any]:
    args: dict[str, Any] = {}
    for key in instance_properties:
        camel_key = to_camel_case(key)
        if camel_key in resource:
            if key in signature.parameters and key in signature.parameters:
                args[key] = _deserialize_value(resource[camel_key], signature.parameters[key])
        elif key in signature.parameters:
            if signature.parameters[key].default is inspect.Parameter.empty:
                args[key] = None
            else:
                args[key] = _deserialize_value(signature.parameters[key].default or None, signature.parameters[key])
    return args


def _load_properties(
    cls: type, resource: dict[str, Any], instance_properties: frozenset[str], signature: inspect.Signature
) -> dict[str, Any]:
    output: dict[str, Any] = {}
    property_by_name = _get_properties_by_name(cls)

    for name, parameter in signature.parameters.items():
        if name in instance_properties:
            continue
        if name in resource:
            output[name] = _deserialize_values(resource[name], parameter)
        elif name in property_by_name:
            property_name = cast(str, property_by_name[name].name)
            if property_name.startswith("__"):
                property_name = property_name[2:]
            if property_name in resource:
                output[name] = _deserialize_values(resource[property_name], parameter)

    return output


def _get_properties_by_name(cls: type) -> dict[str, PropertyOptions]:
    to_search = [cls]
    property_by_name: dict[str, PropertyOptions] = {}
    while to_search:
        current_cls = to_search.pop()
        for name, value in current_cls.__dict__.items():
            if isinstance(value, PropertyOptions):
                property_by_name[name] = value
        to_search.extend(
            [
                b
                for b in current_cls.__bases__
                if {b not in {object, TypedNode, TypedEdge, Edge, Node, TypedNodeApply, TypedEdgeApply}}
            ]
        )
    return property_by_name


_RESERVED_PROPERTY_NAMES = (
    TypedNodeApply._instance_properties
    | TypedEdgeApply._instance_properties
    | TypedNode._instance_properties
    | TypedEdge._instance_properties
)


def _dump_properties(obj: object, camel_case: bool, instance_properties: frozenset[str]) -> dict[str, Any]:
    properties: dict[str, str | int | float | bool | dict | list] = {}
    properties_by_name = _get_properties_by_name(type(obj))
    for key, value in vars(obj).items():
        if key in instance_properties or value is None:
            continue
        if key.startswith("__"):
            key = key[2:]

        if key in properties_by_name:
            key = cast(str, properties_by_name[key].name)

        if isinstance(value, Iterable) and not isinstance(value, (str, dict)):
            properties[key] = [_serialize_property_value(v, camel_case) for v in value]
        else:
            properties[key] = _serialize_property_value(value, camel_case)
    return properties


def _deserialize_values(value: Any, parameter: inspect.Parameter) -> Any:
    if isinstance(value, list):
        return [_deserialize_value(v, parameter) for v in value]
    else:
        return _deserialize_value(value, parameter)


def _deserialize_value(value: Any, parameter: inspect.Parameter) -> Any:
    if parameter.annotation is inspect.Parameter.empty:
        return value
    annotation = str(parameter.annotation)
    if "datetime" in annotation and isinstance(value, str):
        return datetime.fromisoformat(value)
    elif "date" in annotation and isinstance(value, str):
        return date.fromisoformat(value)
    elif DirectRelationReference.__name__ in annotation and isinstance(value, dict):
        return DirectRelationReference.load(value)
    elif NodeId.__name__ in annotation and isinstance(value, dict):
        return NodeId.load(value)
    elif EdgeId.__name__ in annotation and isinstance(value, dict):
        return EdgeId.load(value)

    return value



from __future__ import annotations

from datetime import datetime

from cognite.client.data_classes.data_modeling import DirectRelationReference
from cognite.client.data_classes.data_modeling.ids import ViewId
from cognite.client.data_classes.data_modeling.typed_instances import (
    PropertyOptions,
    TypedEdge,
    TypedEdgeApply,
    TypedNode,
    TypedNodeApply,
)
from cognite.client.utils._experimental import FeaturePreviewWarning

FeaturePreviewWarning("alpha", "alpha", "Core Data Model").warn()


class SourceableProperties:
    source_id = PropertyOptions("sourceId")
    source_created_time = PropertyOptions("sourceCreatedTime")
    source_updated_time = PropertyOptions("sourceUpdatedTime")
    source_created_user = PropertyOptions("sourceCreatedUser")
    source_updated_user = PropertyOptions("sourceUpdatedUser")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Sourceable", "v1")


class SourceableApply(SourceableProperties, TypedNodeApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(space, external_id, existing_version, None, type)
        self.source = source
        self.source_id = source_id
        self.source_created_time = source_created_time
        self.source_updated_time = source_updated_time
        self.source_created_user = source_created_user
        self.source_updated_user = source_updated_user


class Sourceable(
    SourceableProperties,
    TypedNode,
):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        super().__init__(space, external_id, version, last_updated_time, created_time, deleted_time, None, type)
        self.source = source
        self.source_id = source_id
        self.source_created_time = source_created_time
        self.source_updated_time = source_updated_time
        self.source_created_user = source_created_user
        self.source_updated_user = source_updated_user

    def as_write(self) -> SourceableApply:
        return SourceableApply(
            self.space,
            self.external_id,
            self.source_id,
            self.source,
            self.source_created_time,
            self.source_updated_time,
            self.source_created_user,
            self.source_updated_user,
            self.version,
            self.type,
        )


class DescribableProperties:
    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Describable", "v1")


class DescribableApply(DescribableProperties, TypedNodeApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(space, external_id, existing_version, None, type)
        self.name = name
        self.description = description
        self.tags = tags
        self.aliases = aliases


class Describable(
    DescribableProperties,
    TypedNode,
):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        super().__init__(space, external_id, version, last_updated_time, created_time, deleted_time, None, type)
        self.name = name
        self.description = description
        self.tags = tags
        self.aliases = aliases

    def as_write(self) -> DescribableApply:
        return DescribableApply(
            self.space,
            self.external_id,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.version,
            self.type,
        )


class SchedulableProperties:
    start_time = PropertyOptions("startTime")
    end_time = PropertyOptions("endTime")
    scheduled_start_time = PropertyOptions("scheduledStartTime")
    scheduled_end_time = PropertyOptions("scheduledEndTime")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Schedulable", "v1")


class SchedulableApply(SchedulableProperties, TypedNodeApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        start_time: datetime | None = None,
        end_time: datetime | None = None,
        scheduled_start_time: datetime | None = None,
        scheduled_end_time: datetime | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(space, external_id, existing_version, None, type)
        self.start_time = start_time
        self.end_time = end_time
        self.scheduled_start_time = scheduled_start_time
        self.scheduled_end_time = scheduled_end_time


class Schedulable(SchedulableProperties, TypedNode):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        start_time: datetime | None = None,
        end_time: datetime | None = None,
        scheduled_start_time: datetime | None = None,
        scheduled_end_time: datetime | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        super().__init__(space, external_id, version, last_updated_time, created_time, deleted_time, None, type)
        self.start_time = start_time
        self.end_time = end_time
        self.scheduled_start_time = scheduled_start_time
        self.scheduled_end_time = scheduled_end_time

    def as_write(self) -> SchedulableApply:
        return SchedulableApply(
            self.space,
            self.external_id,
            self.start_time,
            self.end_time,
            self.scheduled_start_time,
            self.scheduled_end_time,
            self.version,
            self.type,
        )


class Connection3DProperties:
    revision_id = PropertyOptions("revisionId")
    revision_node_id = PropertyOptions("revisionNodeId")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Connection3D", "v1")


class Connection3DApply(Connection3DProperties, TypedEdgeApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        type: DirectRelationReference | tuple[str, str],
        start_node: DirectRelationReference | tuple[str, str],
        end_node: DirectRelationReference | tuple[str, str],
        revision_id: int,
        revision_node_id: int,
        existing_version: int | None = None,
    ) -> None:
        super().__init__(space, external_id, type, start_node, end_node, existing_version)
        self.revision_id = revision_id
        self.revision_node_id = revision_node_id


class Connection3D(Connection3DProperties, TypedEdge):
    def __init__(
        self,
        space: str,
        external_id: str,
        type: DirectRelationReference | tuple[str, str],
        start_node: DirectRelationReference | tuple[str, str],
        end_node: DirectRelationReference | tuple[str, str],
        version: int,
        last_updated_time: int,
        created_time: int,
        revision_id: int,
        revision_node_id: int,
        deleted_time: int | None = None,
    ) -> None:
        super().__init__(
            space, external_id, version, type, last_updated_time, created_time, start_node, end_node, deleted_time, None
        )
        self.revision_id = revision_id
        self.revision_node_id = revision_node_id

    def as_write(self) -> Connection3DApply:
        return Connection3DApply(
            self.space,
            self.external_id,
            self.type,
            self.start_node,
            self.end_node,
            self.revision_id,
            self.revision_node_id,
            self.version,
        )


class Model3DProperties:
    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Model3D", "v1")


class Model3DApply(Model3DProperties, SourceableApply, DescribableApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        SourceableApply.__init__(
            self,
            space,
            external_id,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            existing_version,
            type,
        )
        DescribableApply.__init__(self, space, external_id, name, description, tags, aliases, existing_version, type)


class Model3D(Model3DProperties, Sourceable, Describable):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        Sourceable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            type,
            deleted_time,
        )
        Describable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            name,
            description,
            tags,
            aliases,
            type,
            deleted_time,
        )

    def as_write(self) -> Model3DApply:
        return Model3DApply(
            self.space,
            self.external_id,
            self.source_id,
            self.source,
            self.source_created_time,
            self.source_updated_time,
            self.source_created_user,
            self.source_updated_user,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.version,
            self.type,
        )


class Object3DProperties:
    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Object3D", "v1")


class Object3DApply(Object3DProperties, SourceableApply, DescribableApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        SourceableApply.__init__(
            self,
            space,
            external_id,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            existing_version,
            type,
        )
        DescribableApply.__init__(self, space, external_id, name, description, tags, aliases, existing_version, type)


class Object3D(Object3DProperties, Sourceable, Describable):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        Sourceable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            type,
            deleted_time,
        )
        Describable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            name,
            description,
            tags,
            aliases,
            type,
            deleted_time,
        )

    def as_write(self) -> Object3DApply:
        return Object3DApply(
            self.space,
            self.external_id,
            self.source_id,
            self.source,
            self.source_created_time,
            self.source_updated_time,
            self.source_created_user,
            self.source_updated_user,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.version,
            self.type,
        )


class AssetTypeProperties:
    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "AssetType", "v1")


class AssetTypeApply(AssetTypeProperties, DescribableApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        code: str | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(space, external_id, name, description, tags, aliases, existing_version, type)
        self.code = code


class AssetType(AssetTypeProperties, Describable):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        code: str | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        super().__init__(
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            name,
            description,
            tags,
            aliases,
            type,
            deleted_time,
        )
        self.code = code

    def as_write(self) -> AssetTypeApply:
        return AssetTypeApply(
            self.space,
            self.external_id,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.code,
            self.version,
            self.type,
        )


class AssetProperties:
    last_path_materialized_time = PropertyOptions("lastPathMaterializedTime")
    asset_type = PropertyOptions("type")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Asset", "v1")


class AssetApply(AssetProperties, Object3DApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        parent: DirectRelationReference | tuple[str, str] | None = None,
        path: DirectRelationReference | tuple[str, str] | None = None,
        last_path_materialization_time: datetime | None = None,
        equipment: DirectRelationReference | tuple[str, str] | None = None,
        asset_type: DirectRelationReference | tuple[str, str] | None = None,
        root: DirectRelationReference | tuple[str, str] | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(
            space,
            external_id,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            name,
            description,
            tags,
            aliases,
            existing_version,
            type,
        )

        self.parent = DirectRelationReference.load(parent) if parent else None
        self.path = DirectRelationReference.load(path) if path else None
        self.last_path_materialization_time = last_path_materialization_time
        self.equipment = DirectRelationReference.load(equipment) if equipment else None
        self.asset_type = DirectRelationReference.load(asset_type) if asset_type else None
        self.root = DirectRelationReference.load(root) if root else None


class Asset(AssetProperties, Object3D):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        parent: DirectRelationReference | tuple[str, str] | None = None,
        path: DirectRelationReference | tuple[str, str] | None = None,
        last_path_materialization_time: datetime | None = None,
        equipment: DirectRelationReference | tuple[str, str] | None = None,
        asset_type: DirectRelationReference | tuple[str, str] | None = None,
        root: DirectRelationReference | tuple[str, str] | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        super().__init__(
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            name,
            description,
            tags,
            aliases,
            type,
            deleted_time,
        )
        self.parent = DirectRelationReference.load(parent) if parent else None
        self.path = DirectRelationReference.load(path) if path else None
        self.last_path_materialization_time = last_path_materialization_time
        self.equipment = DirectRelationReference.load(equipment) if equipment else None
        self.asset_type = DirectRelationReference.load(asset_type) if asset_type else None
        self.root = DirectRelationReference.load(root) if root else None

    def as_write(self) -> AssetApply:
        return AssetApply(
            self.space,
            self.external_id,
            self.parent,
            self.path,
            self.last_path_materialization_time,
            self.equipment,
            self.asset_type,
            self.root,
            self.source_id,
            self.source,
            self.source_created_time,
            self.source_updated_time,
            self.source_created_user,
            self.source_updated_user,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.version,
            self.type,
        )


class EquipmentProperties:
    serial_number = PropertyOptions("serialNumber")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Equipment", "v1")


class EquipmentApply(EquipmentProperties, Object3DApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        serial_number: str | None = None,
        manufacturer: str | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(
            space,
            external_id,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            name,
            description,
            tags,
            aliases,
            existing_version,
            type,
        )
        self.serial_number = serial_number
        self.manufacturer = manufacturer


class Equipment(EquipmentProperties, Object3D):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        serial_number: str | None = None,
        manufacturer: str | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        super().__init__(
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            name,
            description,
            tags,
            aliases,
            type,
            deleted_time,
        )
        self.serial_number = serial_number
        self.manufacturer = manufacturer

    def as_write(self) -> EquipmentApply:
        return EquipmentApply(
            self.space,
            self.external_id,
            self.serial_number,
            self.manufacturer,
            self.source_id,
            self.source,
            self.source_created_time,
            self.source_updated_time,
            self.source_created_user,
            self.source_updated_user,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.version,
            self.type,
        )


class ActivityProperties:
    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "Activity", "v1")


class ActivityApply(ActivityProperties, DescribableApply, SourceableApply, SchedulableApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        assets: list[DirectRelationReference] | list[tuple[str, str]] | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        start_time: datetime | None = None,
        end_time: datetime | None = None,
        scheduled_start_time: datetime | None = None,
        scheduled_end_time: datetime | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        DescribableApply.__init__(self, space, external_id, name, description, tags, aliases, existing_version, type)
        SourceableApply.__init__(
            self,
            space,
            external_id,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            existing_version,
            type,
        )
        SchedulableApply.__init__(
            self,
            space,
            external_id,
            start_time,
            end_time,
            scheduled_start_time,
            scheduled_end_time,
            existing_version,
            type,
        )
        self.assets = [DirectRelationReference.load(asset) for asset in assets] if assets else None


class Activity(ActivityProperties, Describable, Sourceable, Schedulable):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        assets: list[DirectRelationReference] | list[tuple[str, str]] | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        start_time: datetime | None = None,
        end_time: datetime | None = None,
        scheduled_start_time: datetime | None = None,
        scheduled_end_time: datetime | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        Describable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            name,
            description,
            tags,
            aliases,
            type,
            deleted_time,
        )
        Sourceable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            type,
            deleted_time,
        )
        Schedulable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            start_time,
            end_time,
            scheduled_start_time,
            scheduled_end_time,
            type,
            deleted_time,
        )
        self.assets = [DirectRelationReference.load(asset) for asset in assets] if assets else None

    def as_write(self) -> ActivityApply:
        return ActivityApply(
            self.space,
            self.external_id,
            self.assets,
            self.source_id,
            self.source,
            self.source_created_time,
            self.source_updated_time,
            self.source_created_user,
            self.source_updated_user,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.start_time,
            self.end_time,
            self.scheduled_start_time,
            self.scheduled_end_time,
            self.version,
            self.type,
        )


class TimeSeriesProperties:
    is_step = PropertyOptions("isStep")
    is_string = PropertyOptions("isString")
    source_unit = PropertyOptions("sourceUnit")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_cdm_experimental", "TimeSeriesBase", "v1")


class TimesSeriesBaseApply(TimeSeriesProperties, DescribableApply, SourceableApply):
    def __init__(
        self,
        space: str,
        external_id: str,
        is_step: bool,
        is_string: bool,
        source_unit: str | None = None,
        unit: str | None = None,
        assets: list[DirectRelationReference] | list[tuple[str, str]] | None = None,
        equipment: DirectRelationReference | tuple[str, str] | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        DescribableApply.__init__(self, space, external_id, name, description, tags, aliases, existing_version, type)
        SourceableApply.__init__(
            self,
            space,
            external_id,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            existing_version,
            type,
        )
        self.is_step = is_step
        self.is_string = is_string
        self.source_unit = source_unit
        self.unit = unit
        self.assets = [DirectRelationReference.load(asset) for asset in assets] if assets else None
        self.equipment = DirectRelationReference.load(equipment) if equipment else None


class TimeSeriesBase(TimeSeriesProperties, Describable, Sourceable):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        is_step: bool,
        is_string: bool,
        source_unit: str | None = None,
        unit: str | None = None,
        assets: list[DirectRelationReference] | list[tuple[str, str]] | None = None,
        equipment: DirectRelationReference | tuple[str, str] | None = None,
        source_id: str | None = None,
        source: str | None = None,
        source_created_time: datetime | None = None,
        source_updated_time: datetime | None = None,
        source_created_user: str | None = None,
        source_updated_user: str | None = None,
        name: str | None = None,
        description: str | None = None,
        tags: list[str] | None = None,
        aliases: list[str] | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
        deleted_time: int | None = None,
    ) -> None:
        Describable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            name,
            description,
            tags,
            aliases,
            type,
            deleted_time,
        )
        Sourceable.__init__(
            self,
            space,
            external_id,
            version,
            last_updated_time,
            created_time,
            source_id,
            source,
            source_created_time,
            source_updated_time,
            source_created_user,
            source_updated_user,
            type,
            deleted_time,
        )
        self.is_step = is_step
        self.is_string = is_string
        self.source_unit = source_unit
        self.unit = unit
        self.assets = [DirectRelationReference.load(asset) for asset in assets] if assets else None
        self.equipment = DirectRelationReference.load(equipment) if equipment else None

    def as_write(self) -> TimesSeriesBaseApply:
        return TimesSeriesBaseApply(
            self.space,
            self.external_id,
            self.is_step,
            self.is_string,
            self.source_unit,
            self.unit,
            self.assets,
            self.equipment,
            self.source_id,
            self.source,
            self.source_created_time,
            self.source_updated_time,
            self.source_created_user,
            self.source_updated_user,
            self.name,
            self.description,
            self.tags,
            self.aliases,
            self.version,
            self.type,
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import auto
from typing import TYPE_CHECKING, Any, Collection, List, Literal, Union, cast

from typing_extensions import Self, TypeAlias

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    CogniteSort,
    EnumProperty,
    Geometry,
    IdTransformerMixin,
)
from cognite.client.data_classes.aggregations import UniqueResult
from cognite.client.data_classes.labels import Label, LabelDefinition

if TYPE_CHECKING:
    from cognite.client import CogniteClient


class DocumentsGeoJsonGeometry(CogniteObject):
    """Represents the points, curves and surfaces in the coordinate space.

    Args:
        type (Literal["Point", "MultiPoint", "LineString", "MultiLineString", "Polygon", "MultiPolygon", "GeometryCollection"]): The geometry type.
        coordinates (list | None): An array of the coordinates of the geometry. The structure of the elements in this array is determined by the type of geometry.
        geometries (Collection[Geometry] | None): No description.

    Examples:
        Point:
            Coordinates of a point in 2D space, described as an array of 2 numbers.

            Example: `[4.306640625, 60.205710352530346]`

        LineString:
            Coordinates of a line described by a list of two or more points.
            Each point is defined as a pair of two numbers in an array, representing coordinates of a point in 2D space.

            Example: `[[30, 10], [10, 30], [40, 40]]`

        Polygon:
            List of one or more linear rings representing a shape.
            A linear ring is the boundary of a surface or the boundary of a hole in a surface. It is defined as a list consisting of 4 or more Points, where the first and last Point is equivalent.
            Each Point is defined as an array of 2 numbers, representing coordinates of a point in 2D space.

            Example: `[[[35, 10], [45, 45], [15, 40], [10, 20], [35, 10]], [[20, 30], [35, 35], [30, 20], [20, 30]]]`
            type: array

        MultiPoint:
            List of Points. Each Point is defined as an array of 2 numbers, representing coordinates of a point in 2D space.

            Example: `[[35, 10], [45, 45]]`

        MultiLineString:
                List of lines where each line (LineString) is defined as a list of two or more points.
                Each point is defined as a pair of two numbers in an array, representing coordinates of a point in 2D space.

                Example: `[[[30, 10], [10, 30]], [[35, 10], [10, 30], [40, 40]]]`

        MultiPolygon:
            List of multiple polygons.
            Each polygon is defined as a list of one or more linear rings representing a shape.
            A linear ring is the boundary of a surface or the boundary of a hole in a surface. It is defined as a list consisting of 4 or more Points, where the first and last Point is equivalent.
            Each Point is defined as an array of 2 numbers, representing coordinates of a point in 2D space.

            Example: `[[[[30, 20], [45, 40], [10, 40], [30, 20]]], [[[15, 5], [40, 10], [10, 20], [5, 10], [15, 5]]]]`

        GeometryCollection:
            List of geometries as described above.
    """

    _VALID_TYPES = frozenset(
        {"Point", "MultiPoint", "LineString", "MultiLineString", "Polygon", "MultiPolygon", "GeometryCollection"}
    )

    def __init__(
        self,
        type: Literal[
            "Point", "MultiPoint", "LineString", "MultiLineString", "Polygon", "MultiPolygon", "GeometryCollection"
        ],
        coordinates: list | None = None,
        geometries: Collection[Geometry] | None = None,
    ) -> None:
        if type not in self._VALID_TYPES:
            raise ValueError(f"type must be one of {self._VALID_TYPES}")
        self.type = type
        self.coordinates = coordinates
        self.geometries = geometries and list(geometries)

    @classmethod
    def _load(
        cls, raw_geometry: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> DocumentsGeoJsonGeometry:
        instance = cls(
            type=raw_geometry["type"],
            coordinates=raw_geometry.get("coordinates"),
            geometries=raw_geometry.get("geometries"),
        )
        if isinstance(instance.geometries, list):
            instance.geometries = [Geometry.load(geometry) for geometry in instance.geometries]
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.geometries:
            output["geometries"] = [g.dump(camel_case) for g in self.geometries]
        return output


class SourceFile(CogniteObject):
    """
    The source file that a document is derived from.

    Args:
        name (str): The name of the source file.
        hash (str | None): The hash of the source file. This is a SHA256 hash of the original file. The hash only covers the file content, and not other CDF metadata.
        directory (str | None): The directory the file can be found in.
        source (str | None): The source of the file.
        mime_type (str | None): The mime type of the file.
        size (int | None): The size of the file in bytes.
        asset_ids (list[int] | None): The ids of the assets related to this file.
        labels (list[Label | str | LabelDefinition] | None): A list of labels associated with this document's source file in CDF.
        geo_location (DocumentsGeoJsonGeometry | None): The geolocation of the source file.
        dataset_id (int | None): The id if the dataset this file belongs to, if any.
        security_categories (list[int] | None): The security category IDs required to access this file.
        metadata (dict[str, str] | None): Custom, application specific metadata. String key -> String value.
        cognite_client (CogniteClient | None): No description.
        **_ (Any): No description.
    """

    def __init__(
        self,
        name: str,
        hash: str | None = None,
        directory: str | None = None,
        source: str | None = None,
        mime_type: str | None = None,
        size: int | None = None,
        asset_ids: list[int] | None = None,
        labels: list[Label | str | LabelDefinition] | None = None,
        geo_location: DocumentsGeoJsonGeometry | None = None,
        dataset_id: int | None = None,
        security_categories: list[int] | None = None,
        metadata: dict[str, str] | None = None,
        cognite_client: CogniteClient | None = None,
        **_: Any,
    ) -> None:
        self.name = name
        self.hash = hash
        self.directory = directory
        self.source = source
        self.mime_type = mime_type
        self.size = size
        self.asset_ids: list[int] = asset_ids or []
        self.labels: list[Label] = Label._load_list(labels) or []
        self.geo_location = geo_location
        self.dataset_id = dataset_id
        self.security_categories = security_categories
        self.metadata: dict[str, str] = metadata or {}
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> SourceFile:
        return cls(
            name=resource["name"],
            hash=resource.get("hash"),
            directory=resource.get("directory"),
            source=resource.get("source"),
            mime_type=resource.get("mimeType"),
            size=resource.get("size"),
            asset_ids=resource.get("assetIds"),
            labels=Label._load_list(resource.get("labels")),  # type: ignore[arg-type]
            geo_location=DocumentsGeoJsonGeometry._load(resource["geoLocation"]) if "geoLocation" in resource else None,
            dataset_id=resource.get("datasetId"),
            security_categories=resource.get("securityCategories"),
            metadata=resource.get("metadata"),
            cognite_client=cognite_client,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.labels:
            output["labels"] = [label.dump(camel_case) for label in self.labels]
        if self.geo_location:
            output[("geoLocation" if camel_case else "geo_location")] = self.geo_location.dump(camel_case)
        for key in ["metadata", "labels", "asset_ids", "assetIds"]:
            # Remove empty lists and dicts
            if key in output and not output[key]:
                del output[key]
        return output


class Document(CogniteResource):
    """
    A representation of a document in CDF.

    Args:
        id (int): A server-generated ID for the object.
        created_time (int): The creation time of the document in CDF in milliseconds since Jan 1, 1970.
        source_file (SourceFile): The source file that this document is derived from.
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        title (str | None): The title of the document.
        author (str | None): The author of the document.
        producer (str | None): The producer of the document. Many document types contain metadata indicating what software or system was used to create the document.
        modified_time (int | None): The last time the document was modified in CDF in milliseconds since Jan 1, 1970.
        last_indexed_time (int | None): The last time the document was indexed in the search engine, measured in milliseconds since Jan 1, 1970.
        mime_type (str | None): The detected mime type of the document.
        extension (str | None): Extension of the file (always in lowercase)
        page_count (int | None): The number of pages in the document.
        type (str | None): The detected type of the document.
        language (str | None): The detected language of the document.
        truncated_content (str | None): The truncated content of the document.
        asset_ids (list[int] | None): The ids of any assets referred to in the document.
        labels (list[Label | str | LabelDefinition] | None): The labels attached to the document.
        geo_location (DocumentsGeoJsonGeometry | None): The geolocation of the document.
        cognite_client (CogniteClient | None): No description.
        **_ (Any): No description.
    """

    def __init__(
        self,
        id: int,
        created_time: int,
        source_file: SourceFile,
        external_id: str | None = None,
        title: str | None = None,
        author: str | None = None,
        producer: str | None = None,
        modified_time: int | None = None,
        last_indexed_time: int | None = None,
        mime_type: str | None = None,
        extension: str | None = None,
        page_count: int | None = None,
        type: str | None = None,
        language: str | None = None,
        truncated_content: str | None = None,
        asset_ids: list[int] | None = None,
        labels: list[Label | str | LabelDefinition] | None = None,
        geo_location: DocumentsGeoJsonGeometry | None = None,
        cognite_client: CogniteClient | None = None,
        **_: Any,
    ) -> None:
        self.id = id
        self.created_time = created_time
        self.source_file = source_file
        self.external_id = external_id
        self.title = title
        self.author = author
        self.producer = producer
        self.modified_time = modified_time
        self.last_indexed_time = last_indexed_time
        self.mime_type = mime_type
        self.extension = extension
        self.page_count = page_count
        self.type = type
        self.language = language
        self.truncated_content = truncated_content
        self.asset_ids: list[int] = asset_ids or []
        self.labels: list[Label] = Label._load_list(labels) or []
        self.geo_location = geo_location
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Document:
        return cls(
            id=resource["id"],
            created_time=resource["createdTime"],
            source_file=SourceFile._load(resource["sourceFile"]),
            external_id=resource.get("externalId"),
            title=resource.get("title"),
            author=resource.get("author"),
            producer=resource.get("producer"),
            modified_time=resource.get("modifiedTime"),
            last_indexed_time=resource.get("lastIndexedTime"),
            mime_type=resource.get("mimeType"),
            extension=resource.get("extension"),
            page_count=resource.get("pageCount"),
            type=resource.get("type"),
            language=resource.get("language"),
            truncated_content=resource.get("truncatedContent"),
            asset_ids=resource.get("assetIds"),
            labels=Label._load_list(resource.get("labels")),  # type: ignore[arg-type]
            geo_location=DocumentsGeoJsonGeometry._load(resource["geoLocation"]) if "geoLocation" in resource else None,
            cognite_client=cognite_client,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.source_file:
            output[("sourceFile" if camel_case else "source_file")] = self.source_file.dump(camel_case)
        if self.labels:
            output["labels"] = [label.dump(camel_case) for label in self.labels]
        if self.geo_location:
            output[("geoLocation" if camel_case else "geo_location")] = self.geo_location.dump(camel_case)
        return output


class DocumentList(CogniteResourceList[Document], IdTransformerMixin):
    _RESOURCE = Document


@dataclass
class Highlight(CogniteObject):
    """
    Highlighted snippets from name and content fields which show where the query matches are.

    This is used in search results to represent the result.

    Args:
        name (list[str]): Matches in name.
        content (list[str]): Matches in content.
    """

    name: list[str]
    content: list[str]

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "name": self.name,
            "content": self.content,
        }

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        return cls(name=resource["name"], content=resource["content"])


@dataclass
class DocumentHighlight(CogniteResource):
    """
    A pair of a document and highlights.

    This is used in search results to represent the result

    Args:
        highlight (Highlight): The highlight from the document matching search results.
        document (Document): The document.
    """

    highlight: Highlight
    document: Document

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> DocumentHighlight:
        return cls(
            highlight=Highlight._load(resource["highlight"]),
            document=Document._load(resource["document"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {}
        if self.highlight:
            output["highlight"] = self.highlight.dump(camel_case)
        if self.document:
            output["document"] = self.document.dump(camel_case)
        return output


class DocumentHighlightList(CogniteResourceList[DocumentHighlight]):
    _RESOURCE = DocumentHighlight


class DocumentUniqueResult(UniqueResult): ...


class SortableSourceFileProperty(EnumProperty):
    name = auto()  # type: ignore [assignment]
    mime_type = auto()
    source = auto()
    data_set_id = auto()
    metadata = auto()

    def as_reference(self) -> list[str]:
        return ["sourceFile", self.value]


class SourceFileProperty(EnumProperty):
    name = auto()  # type: ignore [assignment]
    mime_type = auto()
    source = auto()
    data_set_id = auto()
    metadata = auto()
    size = auto()
    directory = auto()
    asset_ids = auto()
    asset_external_ids = auto()
    security_categories = auto()
    geo_location = auto()
    labels = auto()

    @staticmethod
    def metadata_key(key: str) -> list[str]:
        return ["sourceFile", "metadata", key]

    def as_reference(self) -> list[str]:
        return ["sourceFile", self.value]


class SortableDocumentProperty(EnumProperty):
    id = auto()
    external_id = auto()
    mime_type = auto()
    extension = auto()
    page_count = auto()
    author = auto()
    title = auto()
    language = auto()
    type = auto()
    created_time = auto()
    modified_time = auto()
    last_indexed_time = auto()


class DocumentProperty(EnumProperty):
    id = auto()
    external_id = auto()
    mime_type = auto()
    extension = auto()
    page_count = auto()
    producer = auto()
    author = auto()
    title = auto()
    language = auto()
    type = auto()
    created_time = auto()
    modified_time = auto()
    last_indexed_time = auto()
    geo_location = auto()
    asset_ids = auto()
    asset_external_ids = auto()
    labels = auto()
    content = auto()


SortableProperty: TypeAlias = Union[SortableSourceFileProperty, SortableDocumentProperty, str, List[str]]


class DocumentSort(CogniteSort):
    def __init__(self, property: SortableProperty, order: Literal["asc", "desc"] = "asc"):
        super().__init__(property, order)


@dataclass
class TemporaryLink:
    url: str
    expires_at: int

    @classmethod
    def load(cls, data: dict[str, Any]) -> TemporaryLink:
        return cls(
            url=data["temporaryLink"],
            expires_at=data["expirationTime"],
        )

    def dump(self) -> dict[str, Any]:
        return {
            "temporaryLink": self.url,
            "expirationTime": self.expires_at,
        }



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal

from typing_extensions import Self

from cognite.client.data_classes._base import CogniteResource
from cognite.client.utils._importing import local_import
from cognite.client.utils._text import convert_all_keys_to_camel_case_recursive, convert_dict_to_case

if TYPE_CHECKING:
    import pandas

    from cognite.client import CogniteClient


class VisionResource(CogniteResource, ABC):
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (bool): Use camelCase for attribute names. Defaults to True.

        Returns:
            dict[str, Any]: A dictionary representation of the instance.
        """
        dumped = {}
        for k, v in vars(self).items():
            if v is None or k.startswith("_"):
                continue
            if isinstance(v, list) and all(isinstance(item, VisionResource) for item in v):
                v = [item.dump(camel_case) for item in v]
            elif isinstance(v, dict) and all(isinstance(item, VisionResource) for item in v.values()):
                v = {item_key: item_value.dump(camel_case) for item_key, item_value in v.items()}
            elif isinstance(v, VisionResource):
                v = v.dump(camel_case)
            dumped[k] = v

        return convert_dict_to_case(dumped, camel_case)

    def to_pandas(self, camel_case: bool = False) -> pandas.DataFrame:  # type: ignore[override]
        """Convert the instance into a pandas DataFrame.

        Args:
            camel_case (bool): Convert column names to camel case (e.g. `externalId` instead of `external_id`)

        Returns:
            pandas.DataFrame: The dataframe.
        """
        pd = local_import("pandas")
        return pd.Series(self.dump(camel_case), name="value").to_frame()


@dataclass
class Point(VisionResource):
    x: float
    y: float

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Point:
        return cls(x=resource["x"], y=resource["y"])


def _process_vertices(vertices: list[dict[str, float]] | list[Point]) -> list[Point]:
    processed_vertices: list[Point] = []
    for v in vertices:
        if isinstance(v, Point):
            processed_vertices.append(v)
        elif isinstance(v, dict) and set(v) == set("xy"):
            processed_vertices.append(Point._load(convert_all_keys_to_camel_case_recursive(v)))
        else:
            raise ValueError(f"{v} is an invalid point.")
    return processed_vertices


@dataclass
class BoundingBox(VisionResource):
    x_min: float
    x_max: float
    y_min: float
    y_max: float

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> BoundingBox:
        return cls(x_min=resource["xMin"], x_max=resource["xMax"], y_min=resource["yMin"], y_max=resource["yMax"])


@dataclass
class CdfResourceRef(VisionResource):
    # A valid reference instance contains exactly one of these
    id: int | None = None
    external_id: str | None = None

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> CdfResourceRef:
        return cls(id=resource.get("id"), external_id=resource.get("externalId"))


@dataclass
class Polygon(VisionResource):
    # A valid polygon contains *at least* three vertices
    vertices: list[Point]

    def __post_init__(self) -> None:
        self.vertices = _process_vertices(self.vertices)

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Polygon:
        return cls(vertices=[Point._load(vertex) for vertex in resource["vertices"]])


@dataclass
class Polyline(VisionResource):
    # A valid polyline contains *at least* two vertices
    vertices: list[Point]

    def __post_init__(self) -> None:
        self.vertices = _process_vertices(self.vertices)

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        return cls(vertices=[Point._load(vertex) for vertex in resource["vertices"]])


@dataclass
class Keypoint(VisionResource):
    point: Point
    confidence: float | None = None

    def __post_init__(self) -> None:
        if isinstance(self.point, dict):
            self.point = Point._load(convert_all_keys_to_camel_case_recursive(self.point))

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Keypoint:
        return cls(point=Point._load(resource["point"]), confidence=resource.get("confidence"))


@dataclass
class Attribute(VisionResource):
    type: Literal["boolean", "numerical"]
    value: bool | float
    description: str | None = None

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Attribute:
        return cls(type=resource["type"], value=resource["value"], description=resource.get("description"))



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, List, TypeVar, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    ExternalIDTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.utils._importing import local_import
from cognite.client.utils._text import convert_all_keys_to_camel_case, to_camel_case, to_snake_case

if TYPE_CHECKING:
    import geopandas

    from cognite.client import CogniteClient

RESERVED_PROPERTIES = {"externalId", "dataSetId", "assetIds", "createdTime", "lastUpdatedTime"}


class FeatureTypeCore(WriteableCogniteResource["FeatureTypeWrite"], ABC):
    """A representation of a feature type in the geospatial API.

    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        data_set_id (int | None): The ID of the dataset this feature type belongs to.
        properties (dict[str, Any] | None): The properties of the feature type.
        search_spec (dict[str, Any] | None): The search spec of the feature type.
    """

    def __init__(
        self,
        external_id: str | None = None,
        data_set_id: int | None = None,
        properties: dict[str, Any] | None = None,
        search_spec: dict[str, Any] | None = None,
    ) -> None:
        self.external_id = external_id
        self.data_set_id = data_set_id
        self.properties = properties
        self.search_spec = search_spec


class FeatureType(FeatureTypeCore):
    """A representation of a feature type in the geospatial API.
    This is the reading version of the FeatureType class, it is used when retrieving feature types from the api.

    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        data_set_id (int | None): The ID of the dataset this feature type belongs to.
        created_time (int | None): The created time of the feature type.
        last_updated_time (int | None): The last updated time of the feature type.
        properties (dict[str, Any] | None): The properties of the feature type.
        search_spec (dict[str, Any] | None): The search spec of the feature type.
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        external_id: str | None = None,
        data_set_id: int | None = None,
        created_time: int | None = None,
        last_updated_time: int | None = None,
        properties: dict[str, Any] | None = None,
        search_spec: dict[str, Any] | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            data_set_id=data_set_id,
            properties=properties,
            search_spec=search_spec,
        )
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> FeatureType:
        return cls(
            external_id=resource.get("externalId"),
            data_set_id=resource.get("dataSetId"),
            created_time=resource.get("createdTime"),
            last_updated_time=resource.get("lastUpdatedTime"),
            properties=resource.get("properties"),
            search_spec=resource.get("searchSpec"),
            cognite_client=cognite_client,
        )

    def as_write(self) -> FeatureTypeWrite:
        """Returns a write version of this feature type."""
        if self.external_id is None or self.properties is None:
            raise ValueError("External ID and properties must be set to create a feature type")

        return FeatureTypeWrite(
            external_id=self.external_id,
            properties=self.properties,
            data_set_id=self.data_set_id,
            search_spec=self.search_spec,
        )


class FeatureTypeWrite(FeatureTypeCore):
    """A representation of a feature type in the geospatial API.
    This is the writing version of the FeatureType class, it is used when creating feature types in the api.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        properties (dict[str, Any]): The properties of the feature type.
        data_set_id (int | None): The ID of the dataset this feature type belongs to.
        search_spec (dict[str, Any] | None): The search spec of the feature type.
    """

    def __init__(
        self,
        external_id: str,
        properties: dict[str, Any],
        data_set_id: int | None = None,
        search_spec: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            data_set_id=data_set_id,
            properties=properties,
            search_spec=search_spec,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> FeatureTypeWrite:
        return cls(
            external_id=resource["externalId"],
            properties=resource["properties"],
            data_set_id=resource.get("dataSetId"),
            search_spec=resource.get("searchSpec"),
        )

    def as_write(self) -> FeatureTypeWrite:
        """Returns this FeatureTypeWrite instance."""
        return self


class FeatureTypeWriteList(CogniteResourceList[FeatureTypeWrite], ExternalIDTransformerMixin):
    _RESOURCE = FeatureTypeWrite


class FeatureTypeList(WriteableCogniteResourceList[FeatureTypeWrite, FeatureType]):
    _RESOURCE = FeatureType

    def as_write(self) -> FeatureTypeWriteList:
        return FeatureTypeWriteList(
            [feature_type.as_write() for feature_type in self], cognite_client=self._get_cognite_client()
        )


@dataclass
class PropertyAndSearchSpec(CogniteObject):
    """A representation of a feature type property and search spec."""

    properties: dict[str, Any] | list[str] | None = None
    search_spec: dict[str, Any] | list[str] | None = None


@dataclass
class Patches(CogniteObject):
    add: dict[str, Any] | None = None
    remove: list[str] | None = None


@dataclass
class FeatureTypePatch(CogniteObject):
    external_id: str | None = None
    property_patches: Patches | None = None
    search_spec_patches: Patches | None = None

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        item = {
            "external_id": self.external_id,
            "property_patches": self.property_patches.dump(camel_case=camel_case) if self.property_patches else None,
            "search_spec_patches": self.search_spec_patches.dump(camel_case=camel_case)
            if self.search_spec_patches
            else None,
        }
        if camel_case:
            return convert_all_keys_to_camel_case(item)
        return item

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        property_patches = (
            Patches.load(resource["propertyPatches"], cognite_client=cognite_client)
            if resource.get("propertyPatches") is not None
            else None
        )
        search_spec_patches = (
            Patches.load(resource["searchSpecPatches"], cognite_client=cognite_client)
            if resource.get("searchSpecPatches") is not None
            else None
        )
        return cls(resource.get("externalId"), property_patches, search_spec_patches)


class FeatureCore(WriteableCogniteResource["FeatureWrite"], ABC):
    """A representation of a feature in the geospatial API.

    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        **properties (Any): The properties of the feature.
    """

    PRE_DEFINED_SNAKE_CASE_NAMES = frozenset({to_snake_case(key) for key in RESERVED_PROPERTIES})

    def __init__(self, external_id: str | None = None, **properties: Any) -> None:
        self.external_id = external_id
        for key in properties:
            setattr(self, key, properties[key])

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        def handle_case(key: str) -> str:
            # Keep properties defined in Feature Type as is
            if camel_case and key in self.PRE_DEFINED_SNAKE_CASE_NAMES:
                return to_camel_case(key)
            return key

        return {
            handle_case(key): value
            for key, value in self.__dict__.items()
            if value is not None and not key.startswith("_")
        }


T_Feature = TypeVar("T_Feature", bound=FeatureCore)


class Feature(FeatureCore):
    """A representation of a feature in the geospatial API.
    This is the reading version of the Feature class, it is used when retrieving features from the api.

    Args:
        external_id (str | None): The external ID provided by the client. Must be unique for the resource type.
        cognite_client (CogniteClient | None): The client to associate with this object.
        **properties (Any): The properties of the feature.
    """

    def __init__(
        self, external_id: str | None = None, cognite_client: CogniteClient | None = None, **properties: Any
    ) -> None:
        super().__init__(external_id=external_id, **properties)
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Feature:
        return cls(
            external_id=resource.get("externalId"),
            cognite_client=cognite_client,
            **{_to_feature_property_name(key): value for key, value in resource.items() if key != "externalId"},
        )

    def as_write(self) -> FeatureWrite:
        """Returns a write version of this feature."""
        if self.external_id is None:
            raise ValueError("External ID must be set to create a feature")

        return FeatureWrite(
            external_id=self.external_id,
            **{
                key: value
                for key, value in self.__dict__.items()
                if value is not None and not key.startswith("_") and key != "external_id"
            },
        )


class FeatureWrite(FeatureCore):
    """A representation of a feature in the geospatial API.
    This is the writing version of the Feature class, it is used when creating features in the api.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        **properties (Any): The properties of the feature.
    """

    def __init__(self, external_id: str, **properties: Any) -> None:
        super().__init__(external_id=external_id, **properties)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> FeatureWrite:
        return cls(
            external_id=resource["externalId"],
            **{_to_feature_property_name(key): value for key, value in resource.items() if key != "externalId"},
        )

    def as_write(self) -> FeatureWrite:
        """Returns this FeatureWrite instance."""
        return self


def _is_geometry_type(property_type: str) -> bool:
    return property_type in {
        "GEOMETRY",
        "POINT",
        "LINESTRING",
        "POLYGON",
        "MULTIPOINT",
        "MULTILINESTRING",
        "MULTIPOLYGON",
        "GEOMETRYCOLLECTION",
        "GEOMETRYZ",
        "POINTZ",
        "LINESTRINGZ",
        "POLYGONZ",
        "MULTIPOINTZ",
        "MULTILINESTRINGZ",
        "MULTIPOLYGONZ",
        "GEOMETRYCOLLECTIONZ",
        "GEOMETRYM",
        "POINTM",
        "LINESTRINGM",
        "POLYGONM",
        "MULTIPOINTM",
        "MULTILINESTRINGM",
        "MULTIPOLYGONM",
        "GEOMETRYCOLLECTIONM",
        "GEOMETRYZM",
        "POINTZM",
        "LINESTRINGZM",
        "POLYGONZM",
        "MULTIPOINTZM",
        "MULTILINESTRINGZM",
        "MULTIPOLYGONZM",
        "GEOMETRYCOLLECTIONZM",
    }


def _to_feature_property_name(property_name: str) -> str:
    return to_snake_case(property_name) if property_name in RESERVED_PROPERTIES else property_name


class FeatureListCore(WriteableCogniteResourceList[FeatureWrite, T_Feature], ExternalIDTransformerMixin):
    def to_geopandas(self, geometry: str, camel_case: bool = False) -> geopandas.GeoDataFrame:
        """Convert the instance into a GeoPandas GeoDataFrame.

        Args:
            geometry (str): The name of the feature type geometry property to use in the GeoDataFrame
            camel_case (bool): Convert column names to camel case (e.g. `externalId` instead of `external_id`)

        Returns:
            geopandas.GeoDataFrame: The GeoPandas GeoDataFrame.

        Examples:

            Convert a FeatureList into a GeoPandas GeoDataFrame:

                >>> from cognite.client.data_classes.geospatial import PropertyAndSearchSpec
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> features = client.geospatial.search_features(...)
                >>> gdf = features.to_geopandas(
                ...     geometry="position",
                ...     camel_case=False
                ... )
                >>> gdf.head()
        """
        df = self.to_pandas(camel_case)
        wkt = local_import("shapely.wkt")
        df[geometry] = df[geometry].apply(lambda g: wkt.loads(g["wkt"]))
        geopandas = local_import("geopandas")
        return geopandas.GeoDataFrame(df, geometry=geometry)

    @staticmethod
    def from_geopandas(
        feature_type: FeatureType,
        geodataframe: geopandas.GeoDataFrame,
        external_id_column: str = "externalId",
        property_column_mapping: dict[str, str] | None = None,
        data_set_id_column: str = "dataSetId",
    ) -> FeatureList:
        """Convert a GeoDataFrame instance into a FeatureList.

        Args:
            feature_type (FeatureType): The feature type the features will conform to
            geodataframe (geopandas.GeoDataFrame): the geodataframe instance to convert into features
            external_id_column (str): the geodataframe column to use for the feature external id
            property_column_mapping (dict[str, str] | None): provides a mapping from featuretype property names to geodataframe columns
            data_set_id_column (str): the geodataframe column to use for the feature dataSet id

        Returns:
            FeatureList: The list of features converted from the geodataframe rows.

        Examples:

            Create features from a geopandas dataframe:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> my_feature_type = ... # some feature type with 'position' and 'temperature' properties
                >>> my_geodataframe = ...  # some geodataframe with 'center_xy', 'temp' and 'id' columns
                >>> feature_list = FeatureList.from_geopandas(feature_type=my_feature_type, geodataframe=my_geodataframe,
                >>>     external_id_column="id", data_set_id_column="dataSetId",
                >>>     property_column_mapping={'position': 'center_xy', 'temperature': 'temp'})
                >>> created_features = client.geospatial.create_features(my_feature_type.external_id, feature_list)

        """
        features = []
        assert feature_type.properties
        if property_column_mapping is None:
            property_column_mapping = {prop_name: prop_name for (prop_name, _) in feature_type.properties.items()}
        for _, row in geodataframe.iterrows():
            feature = Feature(external_id=row[external_id_column], data_set_id=row.get(data_set_id_column, None))
            for prop in feature_type.properties.items():
                prop_name = prop[0]
                # skip generated columns and externalId, dataSetId columns
                if prop_name.startswith("_") or prop_name in [
                    "createdTime",
                    "lastUpdatedTime",
                    "externalId",
                    "dataSetId",
                ]:
                    continue
                prop_type = prop[1]["type"]
                prop_optional = prop[1].get("optional", False)
                column_name = property_column_mapping.get(prop_name, None)
                column_value = row.get(column_name, None)
                column_value = nan_to_none(column_value)
                if column_name is None or column_value is None:
                    if prop_optional:
                        continue
                    else:
                        raise ValueError(f"Missing value for property {prop_name}")

                prop_name = _to_feature_property_name(prop_name)
                if _is_geometry_type(prop_type):
                    setattr(feature, prop_name, {"wkt": column_value.wkt})
                else:
                    setattr(feature, prop_name, column_value)
            features.append(feature)
        return FeatureList(features)


class FeatureWriteList(FeatureListCore[FeatureWrite]):
    _RESOURCE = FeatureWrite

    def as_write(self) -> FeatureWriteList:
        return self


class FeatureList(FeatureListCore[Feature]):
    _RESOURCE = Feature

    def as_write(self) -> FeatureWriteList:
        return FeatureWriteList([feature.as_write() for feature in self], cognite_client=self._get_cognite_client())


def nan_to_none(column_value: Any) -> Any:
    """Convert NaN value to None."""
    from pandas import isna
    from pandas.api.types import is_scalar

    return None if is_scalar(column_value) and isna(column_value) else column_value


class FeatureAggregate(CogniteResource):
    """A result of aggregating features in geospatial API."""

    def __init__(self, cognite_client: CogniteClient | None = None) -> None:
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> FeatureAggregate:
        instance = cls(cognite_client=cognite_client)
        for key, value in resource.items():
            snake_case_key = to_snake_case(key)
            setattr(instance, snake_case_key, value)
        return instance


class FeatureAggregateList(CogniteResourceList[FeatureAggregate]):
    _RESOURCE = FeatureAggregate


class CoordinateReferenceSystemCore(WriteableCogniteResource["CoordinateReferenceSystemWrite"], ABC):
    """A representation of a feature in the geospatial API.

    Args:
        srid (int | None): EPSG code, e.g., 4326. Only valid for geometry types. See https://en.wikipedia.org/wiki/Spatial_reference_system
        wkt (str | None): Well-known text of the geometry, see https://docs.geotools.org/stable/javadocs/org/opengis/referencing/doc-files/WKT.html
        proj_string (str | None): The projection specification string as described in https://proj.org/usage/quickstart.html
    """

    def __init__(
        self,
        srid: int | None = None,
        wkt: str | None = None,
        proj_string: str | None = None,
    ) -> None:
        self.srid = srid
        self.wkt = wkt
        self.proj_string = proj_string


class CoordinateReferenceSystem(CoordinateReferenceSystemCore):
    """A representation of a feature in the geospatial API.
    This is the reading version of the CoordinateReferenceSystem class, it is used when retrieving from the CDF.

    Args:
        srid (int | None): EPSG code, e.g., 4326. Only valid for geometry types. See https://en.wikipedia.org/wiki/Spatial_reference_system
        wkt (str | None): Well-known text of the geometry, see https://docs.geotools.org/stable/javadocs/org/opengis/referencing/doc-files/WKT.html
        proj_string (str | None): The projection specification string as described in https://proj.org/usage/quickstart.html
        cognite_client (CogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        srid: int | None = None,
        wkt: str | None = None,
        proj_string: str | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(
            srid=srid,
            wkt=wkt,
            proj_string=proj_string,
        )
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> CoordinateReferenceSystem:
        return cls(
            srid=resource.get("srid"),
            wkt=resource.get("wkt"),
            proj_string=resource.get("projString"),
            cognite_client=cognite_client,
        )

    def as_write(self) -> CoordinateReferenceSystemWrite:
        """Returns a write version of this coordinate reference system."""
        if self.srid is None or self.wkt is None or self.proj_string is None:
            raise ValueError("SRID, WKT, and projString must be set to create a coordinate reference system")

        return CoordinateReferenceSystemWrite(
            srid=self.srid,
            wkt=self.wkt,
            proj_string=self.proj_string,
        )


class CoordinateReferenceSystemWrite(CoordinateReferenceSystemCore):
    """A representation of a feature in the geospatial API.

    Args:
        srid (int): EPSG code, e.g., 4326. Only valid for geometry types. See https://en.wikipedia.org/wiki/Spatial_reference_system
        wkt (str): Well-known text of the geometry, see https://docs.geotools.org/stable/javadocs/org/opengis/referencing/doc-files/WKT.html
        proj_string (str): The projection specification string as described in https://proj.org/usage/quickstart.html
    """

    def __init__(
        self,
        srid: int,
        wkt: str,
        proj_string: str,
    ) -> None:
        super().__init__(
            srid=srid,
            wkt=wkt,
            proj_string=proj_string,
        )

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None
    ) -> CoordinateReferenceSystemWrite:
        return cls(
            srid=resource["srid"],
            wkt=resource["wkt"],
            proj_string=resource["projString"],
        )

    def as_write(self) -> CoordinateReferenceSystemWrite:
        """Returns this CoordinateReferenceSystemWrite instance."""
        return self


class CoordinateReferenceSystemWriteList(CogniteResourceList[CoordinateReferenceSystemWrite]):
    _RESOURCE = CoordinateReferenceSystemWrite


class CoordinateReferenceSystemList(
    WriteableCogniteResourceList[CoordinateReferenceSystemWrite, CoordinateReferenceSystem]
):
    _RESOURCE = CoordinateReferenceSystem

    def as_write(self) -> CoordinateReferenceSystemWriteList:
        return CoordinateReferenceSystemWriteList(
            [coordinate_reference_system.as_write() for coordinate_reference_system in self],
            cognite_client=self._get_cognite_client(),
        )


class OrderSpec:
    """An order specification with respect to an property."""

    def __init__(self, property: str, direction: str) -> None:
        self.property = property
        self.direction = direction


class RasterMetadata:
    """Raster metadata"""

    def __init__(self, **properties: Any) -> None:
        for key in properties:
            setattr(self, key, properties[key])

    @classmethod
    def load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> RasterMetadata:
        instance = cls(cognite_client=cognite_client)
        for key, value in resource.items():
            snake_case_key = to_snake_case(key)
            setattr(instance, snake_case_key, value)
        return instance


class GeospatialComputeFunction(ABC):
    """A geospatial compute function"""

    @abstractmethod
    def to_json_payload(self) -> dict:
        """Convert function to json for request payload"""


class GeospatialGeometryTransformComputeFunction(GeospatialComputeFunction):
    "A stTransform geospatial compute function"

    def __init__(self, geospatial_geometry_compute_function: GeospatialComputeFunction, srid: int) -> None:
        self.geospatial_geometry_compute_function = geospatial_geometry_compute_function
        self.srid = srid

    def to_json_payload(self) -> dict:
        return {
            "stTransform": {"geometry": self.geospatial_geometry_compute_function.to_json_payload(), "srid": self.srid}
        }


class GeospatialGeometryComputeFunction(GeospatialComputeFunction, ABC):
    "A geospatial geometry compute function"


class GeospatialGeometryValueComputeFunction(GeospatialGeometryComputeFunction):
    """A geospatial geometry value compute function.
    Accepts a well-known text of the geometry prefixed with a spatial reference identifier,
    see https://docs.geotools.org/stable/javadocs/org/opengis/referencing/doc-files/WKT.html
    Args:
        ewkt (str): No description."""

    def __init__(self, ewkt: str) -> None:
        self.ewkt = ewkt

    def to_json_payload(self) -> dict:
        return {
            "ewkt": self.ewkt,
        }


class GeospatialComputedItem(CogniteResource):
    """A representation of an item computed from geospatial."""

    def __init__(self, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> None:
        self.resource = resource
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> GeospatialComputedItem:
        instance = cls(resource=resource, cognite_client=cognite_client)
        for key, value in resource.items():
            snake_case_key = to_snake_case(key)
            setattr(instance, snake_case_key, value)
        return instance


class GeospatialComputedItemList(CogniteResourceList[GeospatialComputedItem]):
    "A list of items computed from geospatial."

    _RESOURCE = GeospatialComputedItem


class GeospatialComputedResponse(CogniteResource):
    "The geospatial compute response."

    def __init__(
        self, computed_item_list: GeospatialComputedItemList, cognite_client: CogniteClient | None = None
    ) -> None:
        self.items = computed_item_list
        self._cognite_client = cast("CogniteClient", cognite_client)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> GeospatialComputedResponse:
        item_list = GeospatialComputedItemList._load(
            cast(List[Any], resource.get("items")), cognite_client=cognite_client
        )
        return cls(item_list, cognite_client=cognite_client)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {"items": self.items.dump(camel_case=camel_case)}
